#!/usr/bin/env python3

from trojanzoo.configs import config
from trojanzoo.datasets import Dataset
from trojanzoo.environ import env
from trojanzoo.utils.fim import KFAC, EKFAC
from trojanzoo.utils.model import (get_all_layer, get_layer, get_layer_name,
                                   activate_params, accuracy, generate_target,
                                   summary)
from trojanzoo.utils.module import get_name, BasicObject
from trojanzoo.utils.output import ansi, prints
from trojanzoo.utils.tensor import add_noise
from trojanzoo.utils.train import train, validate, compare

import torch
import torch.nn as nn
import numpy as np
import os
from collections import OrderedDict
from collections.abc import Iterable

from typing import TYPE_CHECKING
# TODO: python 3.10
from typing import Generator, Iterator, Mapping
from trojanzoo.configs import Config    # TODO: python 3.10
from trojanzoo.utils.model import ExponentialMovingAverage
from torch.optim.optimizer import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
import argparse
from collections.abc import Callable
if TYPE_CHECKING:
    import torch.utils.data

__all__ = ['_Model', 'Model',
           'add_argument', 'create',
           'output_available_models']


class _Model(nn.Module):
    r"""A specific model class which inherits :any:`torch.nn.Module`.

    Args:
        num_classes (int): Number of classes.
        **kwargs: Keyword arguments passed to :meth:`define_preprocess`,
            :meth:`define_features` and :meth:`define_classifier`.

    Attributes:
        num_classes (int): Number of classes. Defaults to ``None``.

        preprocess (torch.nn.Module): Defaults to :meth:`define_preprocess()`.
        features (torch.nn.Module): Defaults to :meth:`define_features()`.
        pool (torch.nn.Module): :any:`torch.nn.AdaptiveAvgPool2d` ``((1, 1))``.
        classifier (torch.nn.Module): Defaults to :meth:`define_classifier()`.
    """

    def __init__(self, num_classes: int = None, **kwargs):
        super().__init__()
        self.preprocess = self.define_preprocess(**kwargs)
        self.features = self.define_features(**kwargs)   # feature extractor
        self.pool = nn.AdaptiveAvgPool2d((1, 1))  # average pooling
        self.flatten = nn.Flatten()
        self.classifier = self.define_classifier(num_classes=num_classes, **kwargs)  # classifier

        self.num_classes = num_classes

    @classmethod
    def define_preprocess(self, **kwargs) -> nn.Module:
        r"""Define preprocess before feature extractor.

        Returns:
            torch.nn.Identity: Identity module.
        """
        return nn.Identity()

    @staticmethod
    def define_features(**kwargs) -> nn.Module:
        r"""Define feature extractor.

        Returns:
            torch.nn.Identity: Identity module.
        """
        return nn.Identity()

    @staticmethod
    def define_classifier(num_features: list[int] = [],
                          num_classes: int = 1000,
                          activation: type[nn.Module] = nn.ReLU,
                          activation_inplace: bool = True,
                          dropout: float = 0.0,
                          **kwargs) -> nn.Sequential:
        r"""
        | Define classifier as
            ``(Linear -> Activation -> Dropout ) * (len(num_features) - 1) -> Linear``.
        | If there is only 1 linear layer, its name will be ``'fc'``.
        | Else, all layer names will be indexed starting from ``0``
            (e.g., ``'fc1', 'relu1', 'dropout0'``).

        Args:
            num_features (list[int]): List of feature numbers.
                Each element serves as the :attr:`in_features` of current layer
                and :attr:`out_features` of preceding layer.
                Defaults to ``[]``.
            num_classes (int): The number of classes.
                This serves as the :attr:`out_features` of last layer.
                Defaults to ``None``.
            activation (type[torch.nn.Module]):
                The type of activation layer.
                Defaults to :any:`torch.nn.ReLU`.
            activation_inplace (bool): Whether to use inplace activation.
                Defaults to ``'True'``
            dropout (float): The drop out probability.
                Will **NOT** add dropout layers if it's ``0``.
                Defaults to ``0.0``.
            **kwargs: Any keyword argument (unused).

        Returns:
            torch.nn.Sequential: The sequential classifier.

        :Examples:
            >>> from trojanzoo.models import _Model
            >>>
            >>> _Model.define_classifier(num_features=[5,4,4], num_classes=10)
            Sequential(
                (fc1): Linear(in_features=5, out_features=4, bias=True)
                (relu1): ReLU(inplace=True)
                (dropout1): Dropout(p=0.5, inplace=False)
                (fc2): Linear(in_features=4, out_features=4, bias=True)
                (relu2): ReLU(inplace=True)
                (dropout2): Dropout(p=0.5, inplace=False)
                (fc3): Linear(in_features=4, out_features=10, bias=True)
            )
        """
        seq = nn.Sequential()
        if len(num_features) == 0:
            return seq
        if activation:
            activation_name = activation.__name__.split('.')[-1].lower()
        if len(num_features) == 1:
            seq.add_module('fc', nn.Linear(num_features[0], num_classes))
        else:
            for i in range(len(num_features) - 1):
                seq.add_module(
                    f'fc{i + 1:d}', nn.Linear(num_features[i], num_features[i + 1]))
                if activation:
                    seq.add_module(f'{activation_name}{i + 1:d}',
                                   activation(inplace=activation_inplace))
                if dropout > 0:
                    seq.add_module(f'dropout{i + 1:d}', nn.Dropout(p=dropout))
            seq.add_module(f'fc{len(num_features):d}', nn.Linear(num_features[-1], num_classes))
        return seq

    def forward(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        r"""``x -> self.get_final_fm -> self.classifier -> return``"""
        x = self.get_final_fm(x, **kwargs)
        x = self.classifier(x)
        return x

    # TODO: combine with get_final_fm ? Consider GNN cases.
    def get_fm(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        r"""``x -> self.preprocess -> self.features -> return``"""
        return self.features(self.preprocess(x))

    def get_final_fm(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        r"""``x -> self.get_fm -> self.pool -> self.flatten -> return``"""
        x = self.get_fm(x, **kwargs)
        x = self.pool(x)
        x = self.flatten(x)
        return x

    def __call__(self, *args, **kwargs) -> torch.Tensor:
        return super().__call__(*args, **kwargs)


class Model(BasicObject):
    r"""
    | A general model wrapper class, which should be the most common interface for users.
    | It inherits :class:`trojanzoo.utils.module.BasicObject`.

    Args:
        name (str): Name of model.
        suffix (str):
            | Suffix of local model weights file (e.g., ``'_adv_train'``).
              Defaults to empty string ``''``.
            | The location of local pretrained weights is
              ``'{folder_path}/{self.name}{self.suffix}.pth'``
        model (type[_Model] | _Model): Type of model or a specific model instance.
        dataset (trojanzoo.datasets.Dataset | None): Corresponding dataset (optional).
            Defaults to ``None``.
        num_classes (int | None): Number of classes.
            If it's ``None``, fetch the value from :attr:`dataset`.
            Defaults to ``None``.
        folder_path (str): Folder path to save model weights.
            Defaults to ``None``.

            Note:
                :attr:`folder_path` is usually
                ``'{model_dir}/{dataset.data_type}/{dataset.name}'``,
                which is claimed as the default value of :func:`create()`.
        official (bool): Whether to use official pretrained weights.
            Defaults to ``False``.
        pretrained (bool): Whether to use local pretrained weights
            from ``'{folder_path}/{self.name}{self.suffix}.pth'``
            Defaults to ``False``.
        randomized_smooth (bool): Whether to use randomized smoothing.
            Defaults to ``False``.
        rs_sigma (float): Randomized smoothing sampling std :math:`\sigma`.
            Defaults to ``0.01``.
        rs_n (int): Randomized smoothing sampling number. Defaults to ``100``.

    Attributes:
        available_models (list[str]): The list of available model names.
        model_urls (dict[str, str]): The links to official pretrained model weights.

        name (str): Name of model.
        suffix (str):
            | Suffix of local model weights file (e.g., ``'_adv_train'``).
              Defaults to empty string ``''``.
            | The location of local pretrained weights is
              ``'{folder_path}/{self.name}{self.suffix}.pth'``
        _model (_Model): :any:`torch.nn.Module` model instance.
        model (torch.nn.DataParallel | _Model):
            Parallel version of :attr:`_model` if there is more than 1 GPU available.
            Generated by :meth:`get_parallel_model()`.
        dataset (trojanzoo.datasets.Dataset | None): Corresponding dataset (optional).
            Defaults to ``None``.
        num_classes (int | None): Number of classes.
            If it's ``None``, fetch the value from :attr:`dataset`.
            Defaults to ``None``.
        folder_path (str): Folder path to save model weights.
            Defaults to ``None``.
        randomized_smooth (bool): Whether to use randomized smoothing.
            Defaults to ``False``.
        rs_sigma (float): Randomized smoothing sampling std :math:`\sigma`.
        rs_n (int): Randomized smoothing sampling number. Defaults to ``100``.

        criterion (~collections.abc.Callable):
            The criterion used to calculate :meth:`loss()`.
        criterion_noreduction (~collections.abc.Callable):
            The criterion used to calculate :meth:`loss()`
            when ``reduction='none'``.
        softmax (torch.nn.Module): :any:`torch.nn.Softmax` ``(dim=1)``.
            Used in :meth:`get_prob()`.
    """

    available_models: list[str] = []
    model_urls: dict[str, str] = []

    @staticmethod
    def add_argument(group: argparse._ArgumentGroup) -> argparse._ArgumentGroup:
        r"""Add model arguments to argument parser group.
        View source to see specific arguments.

        Note:
            This is the implementation of adding arguments.
            The concrete model class may override this method to add more arguments.
            For users, please use :func:`add_argument()` instead, which is more user-friendly.
        """
        group.add_argument('-m', '--model', dest='model_name',
                           help='model name '
                           '(default: config[model][default_model])')
        group.add_argument('--suffix',
                           help='model name suffix (e.g., "_adv_train")')
        group.add_argument('--pretrained', action='store_true',
                           help='load local pretrained weights (default: False)')
        group.add_argument('--official', action='store_true',
                           help='load official pretrained weights (default: False)')
        group.add_argument('--randomized_smooth',
                           help='whether to use randomized smoothing '
                           '(default: False)')
        group.add_argument('--rs_sigma', type=float,
                           help='randomized smoothing sampling std '
                           '(default: 0.01)')
        group.add_argument('--rs_n', type=int,
                           help='randomized smoothing sampling number '
                           '(default: 100)')
        group.add_argument('--model_dir', help='directory to store pretrained models')
        return group

    def __init__(self, name: str = 'model', suffix: str = None,
                 model: type[_Model] | _Model = _Model,
                 dataset: Dataset = None,
                 num_classes: int = None, folder_path: str = None,
                 official: bool = False, pretrained: bool = False,
                 randomized_smooth: bool = False,
                 rs_sigma: float = 0.01, rs_n: int = 100, **kwargs):
        super().__init__()
        self.param_list['model'] = ['folder_path']
        if suffix is not None:
            self.param_list['model'].append('suffix')
        else:
            suffix = ''
        if randomized_smooth:
            self.param_list['model'].extend(
                ['randomized_smooth', 'rs_sigma', 'rs_n'])
        self.name: str = name
        self.dataset = dataset
        self.suffix = suffix
        self.randomized_smooth: bool = randomized_smooth
        self.rs_sigma: float = rs_sigma
        self.rs_n: int = rs_n

        self.folder_path = folder_path
        if folder_path is not None:
            self.folder_path = os.path.normpath(folder_path)
            if not os.path.exists(folder_path):
                os.makedirs(folder_path)

        # ------------Auto-------------- #
        loss_weights: torch.Tensor = kwargs.get('loss_weights', None)
        if dataset:
            if not isinstance(dataset, Dataset):
                raise TypeError(f'{type(dataset)=}    {dataset=}')
            num_classes = num_classes or dataset.num_classes
            loss_weights = kwargs.get('loss_weights', dataset.loss_weights)
        self.num_classes = num_classes  # number of classes
        # TODO: what device shall we save loss_weights?
        # numpy, torch, or torch.cuda.

        match loss_weights:
            case np.ndarray():
                loss_weights = torch.from_numpy(loss_weights).to(device=env['device'], dtype=torch.float)
            case torch.Tensor():
                loss_weights = loss_weights.to(device=env['device'], dtype=torch.float)

        self.loss_weights = loss_weights
        self.layer_name_list: list[str] = None

        # ------------------------------ #
        self.criterion = self.define_criterion(weight=loss_weights)
        self.criterion_noreduction = self.define_criterion(
            weight=loss_weights, reduction='none')
        self.softmax = nn.Softmax(dim=1)
        match model:
            case type():
                if num_classes is not None:
                    kwargs['num_classes'] = num_classes
                self._model = model(name=name, dataset=dataset, **kwargs)
            case nn.Module():
                self._model = model
            case _:
                raise TypeError(type(model))
        self.model = self.get_parallel_model(self._model)
        self.activate_params([])
        if official:
            self.load('official')
        if pretrained:
            self.load(verbose=True)
        self.eval()
        if env['num_gpus']:
            self.cuda()

    # ----------------- Forward Operations ----------------------#

    def get_logits(self, _input: torch.Tensor, parallel: bool = False,
                   randomized_smooth: bool = None,
                   rs_sigma: float = None, rs_n: int = None,
                   **kwargs) -> torch.Tensor:
        r"""Get logits of :attr:`_input`.

        Note:
            Users should use model as Callable function
            rather than call this method directly,
            because ``__call__`` supports :any:`torch.cuda.amp`.

        Args:
            _input (torch.Tensor): The batched input tensor.
            parallel (bool): Whether to use parallel model
                `self.model` rather than `self._model`.
                Defautls to ``False``.
            randomized_smooth (bool | None): Whether to use randomized smoothing.
                If it's ``None``, use :attr:`self.randmized_smooth` instead.
                Defaults to ``None``.
            rs_sigma (float | None): Randomized smoothing sampling std :math:`\sigma`.
                If it's ``None``, use :attr:`self.rs_sigma` instead.
                Defaults to ``None``.
            rs_n (int): Randomized smoothing sampling number.
                If it's ``None``, use :attr:`self.rs_n` instead.
                Defaults to ``None``.
            **kwargs: Keyword arguments passed to :meth:`forward()`.

        Returns:
            torch.Tensor: The logit tensor with shape ``(N, C)``.
        """
        model = self.model if parallel else self._model
        if randomized_smooth is None:
            randomized_smooth = self.randomized_smooth
        if randomized_smooth:
            rs_sigma = rs_sigma if rs_sigma is not None else self.rs_sigma
            rs_n = rs_n if rs_n is not None else self.rs_n
            _list = []
            for _ in range(rs_n):
                # TODO: valid input clip issue
                _input_noise = add_noise(_input, std=rs_sigma)
                _list.append(model(_input_noise, **kwargs))
            return torch.stack(_list).mean(dim=0)
            # TODO: memory issues and parallel possibilities
            # _input_noise = add_noise(repeat_to_batch(
            #     _input, batch_size=n), std=sigma).flatten(end_dim=1)
            # return model(_input_noise, **kwargs).view(
            #     n, len(_input), self.num_classes).mean(dim=0)
        else:
            return model(_input, **kwargs)

    def get_fm(self, _input: torch.Tensor, **kwargs) -> torch.Tensor:
        r"""Get the feature map of :attr:`_input`,
        which is the output of :attr:`self.features`
        and input of :attr:`self.pool`.
        Call :meth:`_Model.get_fm()`.

        Args:
            _input (torch.Tensor): The batched input tensor
                passed to :meth:`_Model.get_fm()`.
            **kwargs: Keyword arguments passed to :meth:`_Model.get_fm()`.

        Returns:
            torch.Tensor: The feature tensor with shape ``(N, C', H', W')``.
        """
        return self._model.get_fm(_input, **kwargs)

    def get_final_fm(self, _input: torch.Tensor, **kwargs) -> torch.Tensor:
        r"""Get the final feature map of :attr:`_input`,
        which is the output of :attr:`self.flatten`
        and input of :attr:`self.classifier`.
        Call :meth:`_Model.get_final_fm()`.

        Args:
            _input (torch.Tensor): The batched input tensor
                passed to :meth:`_Model.get_final_fm()`.
            **kwargs: Keyword arguments passed to :meth:`_Model.get_final_fm()`.

        Returns:
            torch.Tensor: The feature tensor with shape ``(N, dim)``.
        """
        return self._model.get_final_fm(_input, **kwargs)

    def get_prob(self, _input: torch.Tensor, **kwargs) -> torch.Tensor:
        r"""Get the probability classification vector of :attr:`_input`.

        Args:
            _input (torch.Tensor): The batched input tensor
                passed to :meth:`_Model.get_logits()`.
            **kwargs: Keyword arguments passed to :meth:`get_logits()`.

        Returns:
            torch.Tensor: The probability tensor with shape ``(N, C)``.
        """
        return self.softmax(self(_input, **kwargs))

    def get_target_prob(self, _input: torch.Tensor,
                        target: int | list[int] | torch.Tensor,
                        **kwargs) -> torch.Tensor:
        r"""Get the probability w.r.t. :attr:`target` class of :attr:`_input`
        (using :any:`torch.gather`).

        Args:
            _input (torch.Tensor): The batched input tensor
                passed to :meth:`_Model.get_logits()`.
            target (int | list[int] | torch.Tensor): Batched target classes.
            **kwargs: Keyword arguments passed to :meth:`get_logits()`.

        Returns:
            torch.Tensor: The probability tensor with shape ``(N)``.
        """
        match target:
            case int():
                target = [target] * len(_input)
            case list():
                target = torch.tensor(target, device=_input.device)
        return self.get_prob(_input, **kwargs).gather(
            dim=1, index=target.unsqueeze(1)).flatten()

    def get_class(self, _input: torch.Tensor, **kwargs) -> torch.Tensor:
        r"""Get the class classification result of :attr:`_input`
        (using :any:`torch.argmax`).

        Args:
            _input (torch.Tensor): The batched input tensor
                passed to :meth:`_Model.get_logits()`.
            **kwargs: Keyword arguments passed to :meth:`get_logits()`.

        Returns:
            torch.Tensor: The classes tensor with shape ``(N)``.
        """
        return self(_input, **kwargs).argmax(dim=-1)

    def get_layer_name(self, depth: int = -1, prefix: str = '',
                       use_filter: bool = True, non_leaf: bool = False,
                       seq_only: bool = False) -> list[str]:
        r"""Get layer names of model instance.

        Args:
            depth (int): The traverse depth.
                Defaults to ``-1`` (means :math:`\infty`).
            prefix (str): The prefix string to all elements.
                Defaults to empty string ``''``.
            use_filter (bool): Whether to filter out certain layer types.

                * :any:`torchvision.transforms.Normalize`
                * :any:`torch.nn.Dropout`
                * :any:`torch.nn.BatchNorm2d`
                * :any:`torch.nn.ReLU`
                * :any:`torch.nn.Sigmoid`
            non_leaf (bool): Whether to include non-leaf nodes.
                Defaults to ``False``.
            seq_only (bool): Whether to only traverse children
                of :any:`torch.nn.Sequential`.
                If ``False``, will traverse children of all :any:`torch.nn.Module`.
                Defaults to ``False``.

        Returns:
            list[str]: The list of all layer names.

        See Also:
            The implementation is in
            :func:`trojanzoo.utils.model.get_layer_name()`.
        """
        return get_layer_name(self._model, depth, prefix,
                              use_filter, non_leaf, seq_only)

    def get_all_layer(self, _input: torch.Tensor,
                      layer_input: str = 'input', depth: int = -1,
                      prefix='', use_filter: bool = True, non_leaf: bool = False,
                      seq_only: bool = True, verbose: int = 0
                      ) -> dict[str, torch.Tensor]:
        r"""Get all intermediate layer outputs of
        :attr:`_input` from any intermediate layer.

        Args:
            _input (torch.Tensor): The batched input tensor
                from :attr:`layer_input`.
            layer_input (str): The intermediate layer name of :attr:`_input`.
                Defaults to ``'input'``.
            depth (int): The traverse depth.
                Defaults to ``-1`` (:math:`\infty`).
            prefix (str): The prefix string to all elements.
                Defaults to empty string ``''``.
            use_filter (bool): Whether to filter out certain layer types.

                * :any:`torchvision.transforms.Normalize`
                * :any:`torch.nn.Dropout`
                * :any:`torch.nn.BatchNorm2d`
                * :any:`torch.nn.ReLU`
                * :any:`torch.nn.Sigmoid`
            non_leaf (bool): Whether to include non-leaf nodes.
                Defaults to ``False``.
            seq_only (bool): Whether to only traverse children
                of :any:`torch.nn.Sequential`.
                If ``False``, will traverse children of all :any:`torch.nn.Module`.
                Defaults to ``False``.
            verbose (bool): The output level to show information
                including layer name, output shape and module information.
                Setting it larger than ``0`` will enable the output.
                Different integer values stands for different module information.
                Defaults to ``0``.

                * ``0``: No output
                * ``1``: Show layer class name.
                * ``2``: Show layer string (first line).
                * ``3``: Show layer string (full).

        Returns:
            dict[str, torch.Tensor]: The dict of all layer outputs.

        See Also:
            The implementation is in
            :func:`trojanzoo.utils.model.get_all_layer()`.
        """
        return get_all_layer(self._model, _input, layer_input, depth,
                             prefix, use_filter, non_leaf, seq_only, verbose)

    def get_layer(self, _input: torch.Tensor, layer_output: str = 'classifier',
                  layer_input: str = 'input',
                  seq_only: bool = True) -> torch.Tensor:
        r"""Get one certain intermediate layer output
        of :attr:`_input` from any intermediate layer.

        Args:
            _input (torch.Tensor): The batched input tensor
                from :attr:`layer_input`.
            layer_output (str): The intermediate output layer name.
                Defaults to ``'classifier'``.
            layer_input (str): The intermediate layer name of :attr:`_input`.
                Defaults to ``'input'``.
            seq_only (bool): Whether to only traverse children
                of :any:`torch.nn.Sequential`.
                If ``False``, will traverse children of all :any:`torch.nn.Module`.
                Defaults to ``True``.

        Returns:
            torch.Tensor: The output of layer :attr:`layer_output`.

        See Also:
            The implementation is in
            :func:`trojanzoo.utils.model.get_layer()`.
        """
        if layer_input == 'input':
            match layer_output:
                case 'classifier':
                    return self(_input)
                case 'features':
                    return self._model.get_fm(_input)
                case 'flatten':
                    return self.get_final_fm(_input)
        if self.layer_name_list is None:
            self.layer_name_list: list[str] = self.get_layer_name(
                use_filter=False, non_leaf=True)
            self.layer_name_list.insert(0, 'input')
            self.layer_name_list.append('output')
        return get_layer(self._model, _input, layer_output, layer_input,
                         layer_name_list=self.layer_name_list,
                         seq_only=seq_only)

    def loss(self, _input: torch.Tensor = None, _label: torch.Tensor = None,
             _output: torch.Tensor = None, reduction: str = 'mean',
             **kwargs) -> torch.Tensor:
        r"""Calculate the loss using :attr:`self.criterion`
        (:attr:`self.criterion_noreduction`).

        Args:
            _input (torch.Tensor | None): The batched input tensor.
                If :attr:`_output` is provided, this argument will be ignored.
                Defaults to ``None``.
            _label (torch.Tensor): The label of the batch with shape ``(N)``.
            _output (torch.Tensor | None): The logits of :attr:`_input`.
                If ``None``, use :attr:`_input` to calculate logits.
                Defaults to ``None``.
            reduction (str): Specifies the reduction to apply to the output.
                Choose from ``['none', 'mean']``.
                Defaults to ``'mean'``.
            **kwargs: Keyword arguments passed to :meth:`get_logits()`
                if :attr:`_output` is not provided.

        Returns:
            torch.Tensor:
                A scalar loss tensor (with shape ``(N)`` if ``reduction='none'``).
        """
        criterion = self.criterion_noreduction if reduction == 'none' \
            else self.criterion
        if _output is None:
            _output = self(_input, **kwargs)
        return criterion(_output, _label)

    # -------------------------------------------------------- #

    def define_optimizer(
            self, parameters: str | Iterator[nn.Parameter] = 'full',
            OptimType: str | type[Optimizer] = 'SGD',
            lr: float = 0.1, momentum: float = 0.0, weight_decay: float = 0.0,
            lr_scheduler: bool = False,
            lr_scheduler_type: str = 'CosineAnnealingLR',
            lr_step_size: int = 30, lr_gamma: float = 0.1,
            epochs: int = None, lr_min: float = 0.0,
            lr_warmup_epochs: int = 0, lr_warmup_method: str = 'constant',
            lr_warmup_decay: float = 0.01,
            **kwargs) -> tuple[Optimizer, _LRScheduler]:
        r"""Define optimizer and lr_scheduler.

        Args:
            parameters (str | ~collections.abc.Iterable[torch.nn.parameter.Parameter]):
                The parameters to optimize while other model parameters are frozen.
                If :class:`str`, set :attr:`parameters` as:

                    * ``'features': self._model.features``
                    * ``'classifier' | 'partial': self._model.classifier``
                    * ``'full': self._model``

                Defaults to ``'full'``.
            OptimType (str | type[Optimizer]):
                The optimizer type.
                If :class:`str`, load from module :any:`torch.optim`.
                Defaults to ``'SGD'``.
            lr (float): The learning rate of optimizer. Defaults to ``0.1``.
            momentum (float): The momentum of optimizer. Defaults to ``0.0``.
            weight_decay (float): The momentum of optimizer. Defaults to ``0.0``.
            lr_scheduler (bool): Whether to enable lr_scheduler. Defaults to ``False``.
            lr_scheduler_type (str): The type of lr_scheduler.
                Defaults to ``'CosineAnnealingLR'``.

                Available lr_scheduler types (use string rather than type):

                    * :any:`torch.optim.lr_scheduler.StepLR`
                    * :any:`torch.optim.lr_scheduler.CosineAnnealingLR`
                    * :any:`torch.optim.lr_scheduler.ExponentialLR`
            lr_step_size (int): :attr:`step_size` for :any:`torch.optim.lr_scheduler.StepLR`.
                Defaults to ``30``.
            lr_gamma (float): :attr:`gamma` for :any:`torch.optim.lr_scheduler.StepLR`
                or :any:`torch.optim.lr_scheduler.ExponentialLR`.
                Defaults to ``0.1``.
            epochs (int): Total training epochs.
                ``epochs - lr_warmup_epochs`` is passed as :attr:`T_max`
                to any:`torch.optim.lr_scheduler.CosineAnnealingLR`.
                Defaults to ``None``.
            lr_min (float): The minimum of learning rate.
                It's passed as :attr:`eta_min`
                to any:`torch.optim.lr_scheduler.CosineAnnealingLR`.
                Defaults to ``0.0``.
            lr_warmup_epochs (int): Learning rate warmup epochs.
                Passed as :attr:`total_iters` to lr_scheduler.
                Defaults to ``0``.
            lr_warmup_method (str): Learning rate warmup methods.
                Choose from ``['constant', 'linear']``.
                Defaults to ``'constant'``.
            lr_warmup_decay (float): Learning rate warmup decay factor.
                Passed as :attr:`factor` (:attr:`start_factor`) to lr_scheduler.
                Defaults to ``0.01``.
            **kwargs: Keyword arguments passed to optimizer init method.

        Returns:
            (torch.optim.Optimizer, torch.optim.lr_scheduler._LRScheduler):
                The tuple of optimizer and lr_scheduler.
        """
        kwargs['momentum'] = momentum
        kwargs['weight_decay'] = weight_decay
        match parameters:
            case str():
                parameters = self.get_parameter_from_name(name=parameters)
            case Iterable():
                pass
            case _:
                raise TypeError(f'{type(parameters)=}    {parameters=}')
        if isinstance(OptimType, str):
            OptimType: type[Optimizer] = getattr(torch.optim, OptimType)
        keys = OptimType.__init__.__code__.co_varnames
        kwargs = {k: v for k, v in kwargs.items() if k in keys}
        optimizer = OptimType(parameters, lr, **kwargs)
        _lr_scheduler: _LRScheduler = None
        if lr_scheduler:
            main_lr_scheduler: _LRScheduler
            match lr_scheduler_type:
                case 'StepLR':
                    main_lr_scheduler = torch.optim.lr_scheduler.StepLR(
                        optimizer, step_size=lr_step_size, gamma=lr_gamma)
                case 'CosineAnnealingLR':
                    main_lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
                        optimizer, T_max=epochs - lr_warmup_epochs, eta_min=lr_min)
                case 'ExponentialLR':
                    main_lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(
                        optimizer, gamma=lr_gamma)
                case _:
                    raise NotImplementedError(
                        f'Invalid {lr_scheduler_type=}.'
                        'Only "StepLR", "CosineAnnealingLR" and "ExponentialLR" '
                        'are supported.')
            if lr_warmup_epochs > 0:
                match lr_warmup_method:
                    case 'linear':
                        warmup_lr_scheduler = torch.optim.lr_scheduler.LinearLR(
                            optimizer, start_factor=lr_warmup_decay,
                            total_iters=lr_warmup_epochs)
                    case 'constant':
                        warmup_lr_scheduler = torch.optim.lr_scheduler.ConstantLR(
                            optimizer, factor=lr_warmup_decay,
                            total_iters=lr_warmup_epochs)
                    case _:
                        raise NotImplementedError(
                            f'Invalid {lr_warmup_method=}.'
                            'Only "linear" and "constant" are supported.')
                _lr_scheduler = torch.optim.lr_scheduler.SequentialLR(
                    optimizer,
                    schedulers=[warmup_lr_scheduler, main_lr_scheduler],
                    milestones=[lr_warmup_epochs])
            else:
                _lr_scheduler = main_lr_scheduler
        return optimizer, _lr_scheduler

    # TODO: linting, or maybe nn.Module for generic?
    def define_criterion(self, **kwargs) -> nn.CrossEntropyLoss:
        r"""Define criterion to calculate loss.
        Defaults to use :any:`torch.nn.CrossEntropyLoss`.

        Args:
            weight (torch.Tensor | None):
                The loss weights passed to :any:`torch.nn.CrossEntropyLoss`.
                Defaults to :attr:`self.loss_weights`.
            **kwargs: Keyword arguments passed to :any:`torch.nn.CrossEntropyLoss`.
        """
        if 'weight' not in kwargs.keys():
            kwargs['weight'] = self.loss_weights
        return nn.CrossEntropyLoss(**kwargs)
        # if loss_type == 'jsd':
        #     num_classes = num_classes if num_classes is not None \
        #       else self.num_classes

        #     def jsd(_output: torch.Tensor, _label: torch.Tensor, **kwargs):
        #         p: torch.Tensor = F.one_hot(_label, num_classes)
        #         q: torch.Tensor = F.softmax(_output)
        #         log_q = F.log_softmax(_output)
        #         sum_pq = p + q
        #         loss = sum_pq * (sum_pq.log() - math.log(2))
        # return criterion

    # ---------------------Load & Save Model------------------------- #

    @torch.no_grad()
    def load(self, file_path: str = None, folder_path: str = None,
             suffix: str = None, inplace: bool = True,
             map_location: str | Callable | torch.device | dict = 'cpu',
             component: str = 'full', strict: bool = True,
             verbose: bool = False, indent: int = 0,
             **kwargs) -> OrderedDict[str, torch.Tensor]:
        r"""Load pretrained model weights.

        Args:
            file_path (str | None): The file path to load pretrained weights.
                If ``'official'``, call :meth:`get_official_weights()`.
                Defaults to ``'{folder_path}/{self.name}{suffix}.pth'``.
            folder_path (str | None): The folder path containing model checkpoint.
                It is used when :attr:`file_path` is not provided.
                Defaults to :attr:`self.folder_path`.
            suffix (str | None): The suffix string to model weights file.
                Defaults to :attr:`self.suffix`.
            inplace (bool): Whether to change model parameters.
                If ``False``, will only return the dict but not change model parameters.
                Defaults to ``True``.
            map_location (str | ~torch.torch.device | dict):
                Passed to :any:`torch.load`.
                Defaults to ``'cpu'``.

                Note:
                    The device of model parameters will still be ``'cuda'``
                    if there is any cuda available.
                    This argument only affects intermediate operation.
            component (str): Specify which part of the weights to load.
                Choose from ``['full', 'features', 'classifier']``.
                Defaults to ``'full'``.
            strict (bool): Passed to :any:`torch.nn.Module.load_state_dict`.
                Defaults to ``True``.
            verbose (bool): Whether to output auxiliary information.
                Defaults to ``False``.
            indent (int): The indent of output auxialiary information.
            **kwargs: Keyword arguments passed to :any:`torch.load`.

        Returns:
            OrderedDict[str, torch.Tensor]: The model weights OrderedDict.
        """
        map_location = map_location if map_location != 'default' \
            else env['device']
        if file_path is None:
            folder_path = folder_path if folder_path is not None \
                else self.folder_path
            suffix = suffix if suffix is not None else self.suffix
            file_path = os.path.normpath(os.path.join(
                folder_path, f'{self.name}{suffix}.pth'))
        if file_path == 'official':   # TODO
            _dict = self.get_official_weights(map_location=map_location)
            last_bias_value = next(
                reversed(_dict.values()))   # TODO: make sure
            if self.num_classes != len(last_bias_value) \
                    and component != 'features':
                strict = False
                _dict.popitem()
                _dict.popitem()
        else:
            try:
                # TODO: type annotation might change?
                # dict[str, torch.Tensor]
                _dict: OrderedDict[str, torch.Tensor] = torch.load(
                    file_path, map_location=map_location, **kwargs)
            except Exception:
                print(f'{file_path=}')
                raise
        module = self._model
        match component:
            case 'features':
                module = self._model.features
                _dict = OrderedDict(
                    [(key.removeprefix('features.'), value)
                        for key, value in _dict.items()])
            case 'classifier':
                module = self._model.classifier
                _dict = OrderedDict(
                    [(key.removeprefix('classifier.'), value)
                        for key, value in _dict.items()])
            case _:
                assert component == 'full', f'{component=}'
        if inplace:
            try:
                module.load_state_dict(_dict, strict=strict)
            except RuntimeError:
                prints(f'Model {self.name} loaded from: {file_path}',
                       indent=indent)
                raise
        if verbose:
            prints(f'Model {self.name} loaded from: {file_path}',
                   indent=indent)
        if env['num_gpus']:
            self.cuda()
        return _dict

    @torch.no_grad()
    def save(self, file_path: str = None, folder_path: str = None,
             suffix: str = None, component: str = '',
             verbose: bool = False, indent: int = 0, **kwargs):
        r"""Save pretrained model weights.

        Args:
            file_path (str | None): The file path to save pretrained weights.
                Defaults to ``'{folder_path}/{self.name}{suffix}.pth'``.
            folder_path (str | None): The folder path containing model checkpoint.
                It is used when :attr:`file_path` is not provided.
                Defaults to :attr:`self.folder_path`.
            suffix (str | None): The suffix string to model weights file.
                Defaults to :attr:`self.suffix`.
            component (str): Specify which part of the weights to save.
                Choose from ``['full', 'features', 'classifier']``.
                Defaults to ``'full'``.
            verbose (bool): Whether to output auxiliary information.
                Defaults to ``False``.
            indent (int): The indent of output auxialiary information.
            **kwargs: Keyword arguments passed to :any:`torch.save`.
        """
        if file_path is None:
            folder_path = folder_path if folder_path is not None \
                else self.folder_path
            suffix = suffix if suffix is not None else self.suffix
            file_path = os.path.normpath(os.path.join(
                folder_path, f'{self.name}{suffix}.pth'))
        else:
            folder_path = os.path.dirname(file_path)
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
        # TODO: type annotation might change? dict[str, torch.Tensor]
        module = self._model
        if component == 'features':
            module = self._model.features
        elif component == 'classifier':
            module = self._model.classifier
        else:
            assert component == '', f'{component=}'
        _dict: OrderedDict[str, torch.Tensor] = module.state_dict(
            prefix=component)
        torch.save(_dict, file_path, **kwargs)
        if verbose:
            prints(
                f'Model {self.name} saved at: {file_path}', indent=indent)

    def get_official_weights(self, url: str = None,
                             map_location: str | Callable | torch.device | dict = 'cpu',
                             **kwargs) -> OrderedDict[str, torch.Tensor]:
        r"""Get official model weights from :attr:`url`.

        Args:
            url (str | None): The link to model weights.
                Defaults to :attr:`self.model_urls[self.name]`.
            map_location (str | ~torch.torch.device | dict):
                Passed to :any:`torch.hub.load_state_dict_from_url`.
                Defaults to ``'cpu'``.
            **kwargs: Keyword arguments passed to
                :any:`torch.hub.load_state_dict_from_url`.

        Returns:
            OrderedDict[str, torch.Tensor]: The model weights OrderedDict.
        """
        url = self.model_urls[self.name] if url is None else url
        print('get official model weights from: ', url)
        return torch.hub.load_state_dict_from_url(url, map_location=map_location, **kwargs)

    # ---------------------Train and Validate--------------------- #
    # TODO: annotation and remove those arguments to be *args, **kwargs
    def _train(self, epochs: int, optimizer: Optimizer,
               lr_scheduler: _LRScheduler = None,
               lr_warmup_epochs: int = 0,
               model_ema: ExponentialMovingAverage = None,
               model_ema_steps: int = 32,
               grad_clip: float = None, pre_conditioner: None | KFAC | EKFAC = None,
               print_prefix: str = 'Train', start_epoch: int = 0, resume: int = 0,
               validate_interval: int = 10, save: bool = False, amp: bool = False,
               loader_train: torch.utils.data.DataLoader = None,
               loader_valid: torch.utils.data.DataLoader = None,
               epoch_fn: Callable[..., None] = None,
               get_data_fn: Callable[
                   ..., tuple[torch.Tensor, torch.Tensor]] = None,
               loss_fn: Callable[..., torch.Tensor] = None,
               after_loss_fn: Callable[..., None] = None,
               validate_fn: Callable[..., tuple[float, float]] = None,
               save_fn: Callable[..., None] = None, file_path: str = None,
               folder_path: str = None, suffix: str = None,
               writer=None, main_tag: str = 'train', tag: str = '',
               accuracy_fn: Callable[..., list[float]] = None,
               verbose: bool = True, indent: int = 0, **kwargs):
        r"""Train the model"""
        loader_train = loader_train if loader_train is not None \
            else self.dataset.loader['train']
        get_data_fn = get_data_fn if callable(get_data_fn) else self.get_data
        loss_fn = loss_fn if callable(loss_fn) else self.loss
        validate_fn = validate_fn if callable(validate_fn) else self._validate
        save_fn = save_fn if callable(save_fn) else self.save
        accuracy_fn = accuracy_fn if callable(accuracy_fn) else self.accuracy
        kwargs['forward_fn'] = kwargs.get('forward_fn', self.__call__)
        # if not callable(iter_fn) and hasattr(self, 'iter_fn'):
        #     iter_fn = getattr(self, 'iter_fn')
        if not callable(epoch_fn) and hasattr(self, 'epoch_fn'):
            epoch_fn = getattr(self, 'epoch_fn')
        if not callable(after_loss_fn) and hasattr(self, 'after_loss_fn'):
            after_loss_fn = getattr(self, 'after_loss_fn')
        return train(module=self._model, num_classes=self.num_classes,
                     epochs=epochs, optimizer=optimizer, lr_scheduler=lr_scheduler,
                     lr_warmup_epochs=lr_warmup_epochs,
                     model_ema=model_ema, model_ema_steps=model_ema_steps,
                     grad_clip=grad_clip, pre_conditioner=pre_conditioner,
                     print_prefix=print_prefix, start_epoch=start_epoch,
                     resume=resume, validate_interval=validate_interval,
                     save=save, amp=amp,
                     loader_train=loader_train, loader_valid=loader_valid,
                     epoch_fn=epoch_fn, get_data_fn=get_data_fn,
                     loss_fn=loss_fn, after_loss_fn=after_loss_fn,
                     validate_fn=validate_fn,
                     save_fn=save_fn, file_path=file_path,
                     folder_path=folder_path, suffix=suffix,
                     writer=writer, main_tag=main_tag, tag=tag,
                     accuracy_fn=accuracy_fn,
                     verbose=verbose, indent=indent, **kwargs)

    def _validate(self, module: nn.Module = None, num_classes: int = None,
                  loader: torch.utils.data.DataLoader = None,
                  print_prefix: str = 'Validate',
                  indent: int = 0, verbose: bool = True,
                  get_data_fn: Callable[
                      ..., tuple[torch.Tensor, torch.Tensor]] = None,
                  loss_fn: Callable[..., torch.Tensor] = None,
                  writer=None, main_tag: str = 'valid',
                  tag: str = '', _epoch: int = None,
                  accuracy_fn: Callable[..., list[float]] = None,
                  **kwargs) -> tuple[float, float]:
        r"""Evaluate the model.

        Returns:
            (float, float): Accuracy and loss.
        """
        module = self._model if module is None else module
        num_classes = self.num_classes if num_classes is None else num_classes
        loader = loader or self.dataset.loader['valid']
        get_data_fn = get_data_fn or self.get_data
        loss_fn = loss_fn or self.loss
        accuracy_fn = accuracy_fn if callable(accuracy_fn) else self.accuracy
        kwargs['forward_fn'] = kwargs.get('forward_fn', self.__call__)
        return validate(module=module, num_classes=num_classes, loader=loader,
                        print_prefix=print_prefix,
                        indent=indent, verbose=verbose,
                        get_data_fn=get_data_fn,
                        loss_fn=loss_fn,
                        writer=writer, main_tag=main_tag, tag=tag,
                        _epoch=_epoch, accuracy_fn=accuracy_fn, **kwargs)

    def _compare(self, peer: nn.Module = None,
                 loader: torch.utils.data.DataLoader = None,
                 print_prefix: str = 'Validate',
                 indent: int = 0, verbose: bool = True,
                 get_data_fn: Callable[
                     ..., tuple[torch.Tensor, torch.Tensor]] = None,
                 criterion: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = None,
                 **kwargs) -> tuple[float, float]:
        loader = loader or self.dataset.loader['valid']
        get_data_fn = get_data_fn or self.get_data
        criterion = criterion or self.criterion
        return compare(self, peer, loader,
                       print_prefix, indent, verbose,
                       get_data_fn, criterion=self.criterion, **kwargs)

    # ----------------------------Utility--------------------------- #

    def get_data(self, data, **kwargs):
        r"""Process data. Defaults to be :attr:`self.dataset.get_data`.
        If :attr:`self.dataset` is ``None``, return :attr:`data` directly.

        Args:
            data (Any): Unprocessed data.
            **kwargs: Keyword arguments passed to
                :attr:`self.dataset.get_data()`.

        Returns:
            Any: Processed data.
        """
        if self.dataset is not None:
            return self.dataset.get_data(data, **kwargs)
        else:
            return data

    def accuracy(self, _output: torch.Tensor, _label: torch.Tensor,
                 num_classes: int = None,
                 topk: tuple[int] = (1, 5)) -> list[float]:
        r"""Computes the accuracy over the k top predictions
        for the specified values of k.

        Args:
            _output (torch.Tensor): The batched logit tensor with shape ``(N, C)``.
            _label (torch.Tensor): The batched label tensor with shape ``(N)``.
            num_classes (int): Number of classes. Defaults to :attr:`self.num_classes`.
            topk (~collections.abc.Iterable[int]): Which top-k accuracies to show.
                Defaults to ``(1, 5)``.

        Returns:
            list[float]: Top-k accuracies.

        Note:
            The implementation is in :func:`trojanzoo.utils.model.accuracy`.
        """
        num_classes = num_classes or self.num_classes
        return accuracy(_output, _label, num_classes, topk)

    def activate_params(self, params: Iterator[nn.Parameter] = []) -> None:
        r"""Set ``requires_grad=True`` for selected :attr:`params` of :attr:`module`.
        All other params are frozen.

        Args:
            params (~collections.abc.Iterator[torch.nn.parameter.Parameter]):
                The parameters to ``requires_grad``.
                Defaults to ``[]``.
        """
        return activate_params(self._model, params)

    # Need to overload for other packages (GNN)
    # since they are calling their own nn.DataParallel.
    # TODO: nn.parallel.DistributedDataParallel
    @staticmethod
    def get_parallel_model(_model: _Model) -> _Model | nn.DataParallel:
        r"""Get the parallel model if there are more than 1 GPU avaiable.

        Warning:
            :any:`torch.nn.DataParallel` would be deprecated according to
            https://github.com/pytorch/pytorch/issues/65936.
            We need to consider using
            :any:`torch.nn.parallel.DistributedDataParallel` instead.

        Args:
            _model (_Model): The non-parallel model.
        Returns:
            _Model | nn.DataParallel: The parallel model if there are more than 1 GPU avaiable.
        """
        if env['num_gpus'] > 1:
            return nn.DataParallel(_model)
        return _model

    def summary(self, depth: int = None, verbose: bool = True,
                indent: int = 0, **kwargs):
        r"""Prints a string summary of the model instance by calling
        :func:`trojanzoo.utils.module.BasicObject.summary()`
        and :func:`trojanzoo.utils.model.summary()`.

        Args:
            depth (int): Passed to :func:`trojanzoo.utils.model.summary()`.
                If ``None``, set as ``env['verbose']``.
                If still ``None``, set as ``1``.
                Defaults to ``None``.
            verbose (bool): Passed to :func:`trojanzoo.utils.model.summary()`.
                Defaults to ``True``.
            indent (int): Passed to :func:`trojanzoo.utils.module.BasicObject.summary()`
                and passed to :func:`trojanzoo.utils.model.summary()` with ``10`` more.
                Defaults to ``0``.
            **kwargs: Passed to :func:`trojanzoo.utils.model.summary()`.
        """
        super().summary(indent=indent)
        if depth is None:
            depth = env['verbose']
        if depth is None:
            depth = 1
        summary(self._model, depth=depth, verbose=verbose,
                indent=indent + 10, **kwargs)
        prints('-' * 20, indent=indent + 10)

    # -------------------------------Reload---------------------------- #

    def train(self, mode: bool = True):
        r"""Sets the module in training mode.

        See Also:
            :any:`torch.nn.Module.train`.
        """
        self._model.train(mode=mode)
        self.model.train(mode=mode)
        return self

    def eval(self):
        r"""Sets the module in evaluation mode.

        See Also:
            :any:`torch.nn.Module.eval`.
        """
        self._model.eval()
        self.model.eval()
        return self

    def cpu(self):
        r"""Moves all model parameters and buffers to the CPU.

        See Also:
            :any:`torch.nn.Module.cpu`.
        """
        self._model.cpu()
        self.model.cpu()
        return self

    def cuda(self, device: None | int | torch.device = None):
        r"""Moves all model parameters and buffers to the GPU.

        See Also:
            :any:`torch.nn.Module.cuda`.
        """
        self._model.cuda(device=device)
        self.model.cuda(device=device)
        return self

    def zero_grad(self, set_to_none: bool = False):
        r"""Sets gradients of all model parameters to zero.

        See Also:
            :any:`torch.nn.Module.zero_grad`.
        """
        return self._model.zero_grad(set_to_none=set_to_none)

    def state_dict(self, destination: Mapping[str, torch.Tensor] = None,
                   prefix: str = '', keep_vars: bool = False):
        r"""Returns a dictionary containing a whole state of the module.

        See Also:
            :any:`torch.nn.Module.state_dict`.
        """
        return self._model.state_dict(destination=destination, prefix=prefix,
                                      keep_vars=keep_vars)

    def load_state_dict(self, state_dict: dict[str, torch.Tensor],
                        strict: bool = True):
        r"""Copies parameters and buffers from :attr:`state_dict`
        into this module and its descendants.

        See Also:
            :any:`torch.nn.Module.load_state_dict`.
        """
        return self._model.load_state_dict(state_dict, strict=strict)

    def parameters(self, recurse: bool = True):
        r"""Returns an iterator over module parameters.

        See Also:
            :any:`torch.nn.Module.parameters`.
        """
        return self._model.parameters(recurse=recurse)

    def named_parameters(self, prefix='', recurse=True):
        r"""Returns an iterator over module parameters, yielding both the
        name of the parameter as well as the parameter itself.

        See Also:
            :any:`torch.nn.Module.named_parameters`.
        """
        return self._model.named_parameters(prefix=prefix, recurse=recurse)

    def children(self):
        r"""Returns an iterator over immediate children modules.

        See Also:
            :any:`torch.nn.Module.children`.
        """
        return self._model.children()

    def named_children(self):
        r"""Returns an iterator over immediate children modules, yielding both
        the name of the module as well as the module itself.

        See Also:
            :any:`torch.nn.Module.named_children`.
        """
        return self._model.named_children()

    def modules(self):
        r"""Returns an iterator over all modules in the network.

        See Also:
            :any:`torch.nn.Module.modules`.
        """
        return self._model.modules()

    def named_modules(self, memo: None | set[nn.Module] = None,
                      prefix: str = ''
                      ) -> Generator[tuple[str, nn.Module], None, None]:
        r"""Returns an iterator over all modules in the network, yielding
        both the name of the module as well as the module itself.

        See Also:
            :any:`torch.nn.Module.named_modules`.
        """
        return self._model.named_modules(memo=memo, prefix=prefix)

    def apply(self, fn: Callable[['nn.Module'], None]):
        r"""Applies ``fn`` recursively to every submodule (as returned by ``.children()``)
        as well as self. Typical use includes initializing the parameters of a model

        See Also:
            :any:`torch.nn.Module.apply`.
        """
        return self._model.apply(fn)

    def requires_grad_(self, requires_grad: bool = True):
        r"""Change if autograd should record operations on parameters in this
        module.

        See Also:
            :any:`torch.nn.Module.requires_grad_`.
        """
        return self._model.requires_grad_(requires_grad=requires_grad)

    # ----------------------------------------------------------------- #

    @torch.no_grad()
    def remove_misclassify(self, data: tuple[torch.Tensor, torch.Tensor],
                           **kwargs):
        r"""Remove misclassified samples in a data batch.

        Args:
            data (tuple[torch.Tensor, torch.Tensor]):
                The input and label to process with shape ``(N, *)`` and ``(N)``.
            **kwargs: Keyword arguments passed to :meth:`get_data`.

        Returns:
            (torch.Tensor, torch.Tensor):
                The processed input and label with shape ``(N - k, *)`` and ``(N - k)``.
        """
        _input, _label = self.get_data(data, **kwargs)
        _classification = self.get_class(_input)
        repeat_idx = _classification.eq(_label)
        return _input[repeat_idx], _label[repeat_idx]

    def generate_target(self, _input: torch.Tensor, idx: int = 1,
                        same: bool = False) -> torch.Tensor:
        r"""Generate target labels of a batched input based on
            the classification confidence ranking index.

        Args:
            _input (torch.Tensor): The input tensor.
            idx (int): The classification confidence
                rank of target class.
                Defaults to ``1``.
            same (bool): Generate the same label
                for all samples using mod.
                Defaults to ``False``.

        Returns:
            torch.Tensor:
                The generated target label with shape ``(N)``.

        See Also:
            The implementation is in
            :func:`trojanzoo.utils.model.generate_target()`.
        """
        return generate_target(self, _input, idx, same)

    # --------------------- Undocumented Methods ------------------------- #

    def get_parameter_from_name(self, name: str = 'full'
                                ) -> Iterator[nn.Parameter]:
        match name:
            case 'features':
                params = self._model.features.parameters()
            case 'classifier' | 'partial':
                params = self._model.classifier.parameters()
            case 'full':
                params = self._model.parameters()
            case _:
                raise NotImplementedError(f'{name=}')
        return params

    def __call__(self, _input: torch.Tensor, amp: bool = False,
                 **kwargs) -> torch.Tensor:
        if amp:
            with torch.cuda.amp.autocast():
                return self.get_logits(_input, **kwargs)
        return self.get_logits(_input, **kwargs)


def add_argument(parser: argparse.ArgumentParser,
                 model_name: None | str = None,
                 model: None | str | Model = None,
                 config: Config = config,
                 class_dict: dict[str, type[Model]] = {}
                 ) -> argparse._ArgumentGroup:
    r"""
    | Add model arguments to argument parser.
    | For specific arguments implementation, see :meth:`Model.add_argument()`.

    Args:
        parser (argparse.ArgumentParser): The parser to add arguments.
        model_name (str): The model name.
        model (str | Model): The model instance or model name
            (as the alias of `model_name`).
        config (Config): The default parameter config,
            which contains the default dataset and model name if not provided.
        class_dict (dict[str, type[Model]]):
            Map from model name to model class.
            Defaults to ``{}``.

    Returns:
        argparse._ArgumentGroup: The argument group.
    """
    dataset_name = get_name(arg_list=['-d', '--dataset'])
    if dataset_name is None:
        dataset_name = config.full_config['dataset']['default_dataset']
    model_name = get_name(name=model_name, module=model,
                          arg_list=['-m', '--model'])
    if model_name is None:
        model_name = config.get_config(dataset_name=dataset_name)[
            'model']['default_model']
    model_name = get_model_class(model_name, class_dict=class_dict)

    group = parser.add_argument_group(
        '{yellow}model{reset}'.format(**ansi), description=model_name)
    model_class_name = get_model_class(model_name, class_dict=class_dict)
    try:
        ModelType = class_dict[model_class_name]
    except KeyError:
        print(f'{model_class_name} not in \n{list(class_dict.keys())}')
        raise
    return ModelType.add_argument(group)


def create(model_name: None | str = None, model: None | str | Model = None,
           dataset_name: None | str = None, dataset: None | str | Dataset = None,
           config: Config = config,
           class_dict: dict[str, type[Model]] = {},
           **kwargs) -> Model:
    r"""
    | Create a model instance.
    | For arguments not included in :attr:`kwargs`,
      use the default values in :attr:`config`.
    | The default value of :attr:`folder_path` is
      ``'{model_dir}/{dataset.data_type}/{dataset.name}'``.
    | For model implementation, see :class:`Model`.

    Args:
        model_name (str): The model name.
        model (str | Model): The model instance or model name
            (as the alias of `model_name`).
        dataset_name (str): The dataset name.
        dataset (str | trojanzoo.datasets.Dataset):
            Dataset instance or dataset name
            (as the alias of `dataset_name`).
        config (Config): The default parameter config.
        class_dict (dict[str, type[Model]]):
            Map from model name to model class.
            Defaults to ``{}``.
        **kwargs: The keyword arguments
            passed to model init method.

    Returns:
        Model: The model instance.
    """
    dataset_name = get_name(
        name=dataset_name, module=dataset, arg_list=['-d', '--dataset'])
    model_name = get_name(name=model_name, module=model,
                          arg_list=['-m', '--model'])
    if dataset_name is None:
        dataset_name = config.full_config['dataset']['default_dataset']
    if model_name is None:
        model_name = config.get_config(dataset_name=dataset_name)[
            'model']['default_model']
    result = config.get_config(dataset_name=dataset_name)[
        'model'].update(kwargs)
    model_name = model_name if model_name is not None \
        else result['default_model']

    name_list = [name for sub_list in get_available_models(
        class_dict=class_dict).values()
        for name in sub_list]
    name_list = sorted(name_list)
    assert model_name in name_list, f'{model_name} not in \n{name_list}'
    model_class_name = get_model_class(model_name, class_dict=class_dict)
    try:
        ModelType = class_dict[model_class_name]
    except KeyError:
        print(f'{model_class_name} not in \n{list(class_dict.keys())}')
        raise

    if 'folder_path' not in result.keys() and isinstance(dataset, Dataset):
        result['folder_path'] = os.path.join(result['model_dir'],
                                             dataset.data_type,
                                             dataset.name)
    return ModelType(name=model_name, dataset=dataset, dataset_name=dataset_name, **result)


def output_available_models(class_dict: dict[str, type[Model]] = {},
                            indent: int = 0) -> None:
    r"""Output all available model names.

    Args:
        class_dict (dict[str, type[Model]]): Map from model name to model class.
            Defaults to ``{}``.
        indent (int): The space indent for the entire string.
            Defaults to ``0``.
    """
    names_dict = get_available_models(class_dict)
    for k in sorted(names_dict.keys()):
        prints('{yellow}{k}{reset}'.format(k=k, **ansi), indent=indent)
        prints(names_dict[k], indent=indent + 10)
        print()


def get_available_models(class_dict: dict[str, type[Model]] = {}
                         ) -> dict[str, list[str]]:
    return {k: v.available_models for k, v in class_dict.items()}


def get_model_class(name: str, class_dict: dict[str, type[Model]] = {}) -> str:
    correct_name: str = None
    for class_name in class_dict.keys():
        if class_name in name.lower() and \
                (correct_name is None or
                 len(class_name) > len(correct_name)):
            correct_name = class_name
    if correct_name is not None:
        return correct_name
    raise KeyError(f'{name} not in {list(class_dict.keys())}')



#!/usr/bin/env python3

from trojanzoo.attacks import Attack

from trojanvision.datasets.imageset import ImageSet
from trojanvision.models.imagemodel import ImageModel
from trojanvision.marks import Watermark
from trojanzoo.environ import env
from trojanzoo.utils.data import TensorListDataset, sample_batch
from trojanzoo.utils.logger import SmoothedValue


import torch
import torchvision.transforms.functional as F
import numpy as np
import functools
import math
import random
import os

import argparse
from collections.abc import Callable
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import torch.utils.data


class BackdoorAttack(Attack):
    r"""Backdoor attack abstract class.
    It inherits :class:`trojanzoo.attacks.Attack`.

    Note:
        This class is actually equivalent to :class:`trojanvision.attacks.BadNet`.

    BackdoorAttack attaches a provided watermark to some training images
    and inject them into training set with target label.
    After retraining, the model will classify images with watermark
    of certain/all classes into target class.

    Args:
        mark (trojanvision.marks.Watermark): The watermark instance.
        target_class (int): The target class that images with watermark will be misclassified as.
            Defaults to ``0``.
        poison_percent (float): Percentage of poisoning inputs in the whole training set.
            Defaults to ``0.01``.
        train_mode (float): Training mode to inject backdoor.
            Choose from ``['batch', 'dataset', 'loss']``.
            Defaults to ``'batch'``.

            * ``'batch'``: For a clean batch, randomly picked :attr:`poison_num` inputs,
              attach trigger on them, modify their labels and append to original batch.
            * ``'dataset'``: Create a poisoned dataset and use the mixed dataset.
            * ``'loss'``: For a clean batch, calculate the loss on clean data
              and the loss on poisoned data (all batch)
              and mix them using :attr:`poison_percent` as weight.

    Attributes:
        poison_ratio (float): The ratio of poison data divided by clean data.
            ``poison_percent / (1 - poison_percent)``
        poison_num (float | int): The number of poison data in each batch / dataset.

            * ``train_mode == 'batch'  : poison_ratio * batch_size``
            * ``train_mode == 'dataset': int(poison_ratio * len(train_set))``
            * ``train_mode == 'loss'   : N/A``
        poison_set (torch.utils.data.Dataset):
            Poison dataset (no clean data) ``if train_mode == 'dataset'``.
    """

    name: str = 'backdoor_attack'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--target_class', type=int,
                           help='target class of backdoor '
                           '(default: 0)')
        group.add_argument('--poison_percent', type=float,
                           help='malicious training data proportion '
                           '(default: 0.01)')
        group.add_argument('--train_mode', choices=['batch', 'dataset', 'loss'],
                           help='training mode to inject backdoor '
                           '(default: "batch")')
        return group

    def __init__(self, mark: Watermark = None,
                 source_class: list[int] = None,
                 target_class: int = 0, poison_percent: float = 0.01,
                 train_mode: str = 'batch', **kwargs):
        super().__init__(**kwargs)
        self.dataset: ImageSet
        self.model: ImageModel
        self.mark = mark
        self.param_list['backdoor'] = ['train_mode', 'target_class', 'poison_percent', 'poison_num']
        self.source_class = source_class
        self.target_class = target_class
        self.poison_percent = poison_percent
        self.poison_ratio = self.poison_percent / (1 - self.poison_percent)
        self.train_mode = train_mode
        match train_mode:
            case 'batch':
                self.poison_num = self.dataset.batch_size * self.poison_ratio
                self.poison_set = None
            case 'dataset':
                self.poison_num = int(len(self.dataset.loader['train'].dataset) * self.poison_ratio)
                self.poison_set = self.get_poison_dataset()
            case _:
                self.poison_set = None

    def attack(self, epochs: int, **kwargs):
        match self.train_mode:
            case 'batch':
                loader = self.dataset.get_dataloader(
                    'train', batch_size=self.dataset.batch_size + int(self.poison_num))
                return self.model._train(epochs, loader_train=loader,
                                         validate_fn=self.validate_fn,
                                         get_data_fn=self.get_data,
                                         save_fn=self.save, **kwargs)
            case 'dataset':
                mix_dataset = torch.utils.data.ConcatDataset([self.dataset.loader['train'].dataset,
                                                              self.poison_set])
                loader = self.dataset.get_dataloader('train', dataset=mix_dataset)
                return self.model._train(epochs, loader_train=loader,
                                         validate_fn=self.validate_fn,
                                         save_fn=self.save, **kwargs)
            case 'loss':
                if 'loss_fn' in kwargs.keys():
                    kwargs['loss_fn'] = functools.partial(self.loss_weighted, loss_fn=kwargs['loss_fn'])
                else:
                    kwargs['loss_fn'] = self.loss_weighted
                return self.model._train(epochs,
                                         validate_fn=self.validate_fn,
                                         save_fn=self.save, **kwargs)
            case _:
                raise NotImplementedError(f'{self.train_mode=}')

    def get_poison_dataset(self, poison_label: bool = True,
                           poison_num: int = None,
                           seed: int = None
                           ) -> torch.utils.data.Dataset:
        r"""Get poison dataset (no clean data).

        Args:
            poison_label (bool):
                Whether to use target poison label for poison data.
                Defaults to ``True``.
            poison_num (int): Number of poison data.
                Defaults to ``round(self.poison_ratio * len(train_set))``
            seed (int): Random seed to sample poison input indices.
                Defaults to ``env['data_seed']``.

        Returns:
            torch.utils.data.Dataset:
                Poison dataset (no clean data).
        """
        if seed is None:
            seed = env['data_seed']
        torch.random.manual_seed(seed)
        train_set = self.dataset.loader['train'].dataset
        poison_num = poison_num or round(self.poison_ratio * len(train_set))
        _input, _label = sample_batch(train_set, batch_size=poison_num)
        _label = _label.tolist()

        if poison_label:
            _label = [self.target_class] * len(_label)
        trigger_input = self.add_mark(_input)
        return TensorListDataset(trigger_input, _label)

    def get_filename(self, mark_alpha: float = None, target_class: int = None, **kwargs) -> str:
        r"""Get filenames for current attack settings."""
        if mark_alpha is None:
            mark_alpha = self.mark.mark_alpha
        if target_class is None:
            target_class = self.target_class
        mark_filename = os.path.split(self.mark.mark_path)[-1]
        mark_name, mark_ext = os.path.splitext(mark_filename)
        _file = '{mark}_tar{target:d}_alpha{mark_alpha:.2f}_mark({mark_height:d},{mark_width:d})'.format(
            mark=mark_name, target=target_class, mark_alpha=mark_alpha,
            mark_height=self.mark.mark_height, mark_width=self.mark.mark_width)
        if self.mark.mark_random_pos:
            _file = 'randompos_' + _file
        if self.mark.mark_scattered:
            _file = 'scattered_' + _file
        return _file

    # ---------------------- I/O ----------------------------- #

    def save(self, filename: str = None, **kwargs):
        r"""Save attack results to files."""
        filename = filename or self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        np.save(file_path + '.npy', self.mark.mark.detach().cpu().numpy())
        F.to_pil_image(self.mark.mark).save(file_path + '.png')
        self.model.save(file_path + '.pth')
        print('attack results saved at: ', file_path)

    def load(self, filename: str = None, **kwargs):
        r"""Load attack results from previously saved files."""
        filename = filename or self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        self.mark.load_mark(file_path + '.npy', already_processed=True)
        self.model.load(file_path + '.pth')
        print('attack results loaded from: ', file_path)

    # ---------------------- Utils ---------------------------- #

    def add_mark(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        r"""Add watermark to input tensor.
        Defaults to :meth:`trojanvision.marks.Watermark.add_mark()`.
        """
        return self.mark.add_mark(x, **kwargs)

    def loss_weighted(self, _input: torch.Tensor = None, _label: torch.Tensor = None,
                      _output: torch.Tensor = None, loss_fn: Callable[..., torch.Tensor] = None,
                      **kwargs) -> torch.Tensor:
        loss_fn = loss_fn if loss_fn is not None else self.model.loss
        loss_clean = loss_fn(_input, _label, **kwargs)
        trigger_input = self.add_mark(_input)
        trigger_label = self.target_class * torch.ones_like(_label)
        loss_poison = loss_fn(trigger_input, trigger_label, **kwargs)
        return (1 - self.poison_percent) * loss_clean + self.poison_percent * loss_poison

    def get_data(self, data: tuple[torch.Tensor, torch.Tensor],
                 org: bool = False, keep_org: bool = True,
                 poison_label: bool = True, **kwargs
                 ) -> tuple[torch.Tensor, torch.Tensor]:
        r"""Get data.

        Args:
            data (tuple[torch.Tensor, torch.Tensor]): Tuple of input and label tensors.
            org (bool): Whether to return original clean data directly.
                Defaults to ``False``.
            keep_org (bool): Whether to keep original clean data in final results.
                If ``False``, the results are all infected.
                Defaults to ``True``.
            poison_label (bool): Whether to use target class label for poison data.
                Defaults to ``True``.
            **kwargs: Any keyword argument (unused).

        Returns:
            (torch.Tensor, torch.Tensor): Result tuple of input and label tensors.
        """
        _input, _label = self.model.get_data(data)
        if not org:
            if keep_org:
                decimal, integer = math.modf(len(_label) * self.poison_ratio)
                integer = int(integer)
                if random.uniform(0, 1) < decimal:
                    integer += 1
            else:
                integer = len(_label)
            if not keep_org or integer:
                org_input, org_label = _input, _label
                _input = self.add_mark(org_input[:integer])
                _label = _label[:integer]
                if poison_label:
                    _label = self.target_class * torch.ones_like(org_label[:integer])
                if keep_org:
                    _input = torch.cat((_input, org_input))
                    _label = torch.cat((_label, org_label))
        return _input, _label

    def validate_fn(self,
                    get_data_fn: Callable[..., tuple[torch.Tensor, torch.Tensor]] = None,
                    loss_fn: Callable[..., torch.Tensor] = None,
                    main_tag: str = 'valid', indent: int = 0,
                    threshold: float = 5.0,
                    **kwargs) -> tuple[float, float]:
        clean_acc, _ = self.model._validate(print_prefix='Validate Clean', main_tag='valid clean',
                                            get_data_fn=None, indent=indent, **kwargs)
        asr, _ = self.model._validate(print_prefix='Validate ASR', main_tag='valid asr',
                                      get_data_fn=self.get_data, keep_org=False, poison_label=True,
                                      indent=indent, **kwargs)
        # self.model._validate(print_prefix='Validate Trigger Org', main_tag='',
        #                      get_data_fn=self.get_data, keep_org=False, poison_label=False,
        #                      indent=indent, **kwargs)
        # prints(f'Validate Confidence: {self.validate_confidence():.3f}', indent=indent)
        # prints(f'Neuron Jaccard Idx: {self.get_neuron_jaccard():.3f}', indent=indent)
        if self.clean_acc - clean_acc > threshold:
            asr = 0.0
        return asr, clean_acc

    @torch.no_grad()
    def validate_confidence(self, mode: str = 'valid', success_only: bool = True) -> float:
        r"""Get :attr:`self.target_class` confidence on dataset of :attr:`mode`.

        Args:
            mode (str): Dataset mode. Defaults to ``'valid'``.
            success_only (bool): Whether to only measure confidence
                on attack-successful inputs.
                Defaults to ``True``.

        Returns:
            float: Average confidence of :attr:`self.target_class`.
        """
        source_class = self.source_class or list(range(self.dataset.num_classes))
        source_class = source_class.copy()
        if self.target_class in source_class:
            source_class.remove(self.target_class)
        loader = self.dataset.get_dataloader(mode=mode, class_list=source_class)

        confidence = SmoothedValue()
        for data in loader:
            _input, _label = self.model.get_data(data)
            trigger_input = self.add_mark(_input)
            trigger_label = self.model.get_class(trigger_input)
            if success_only:
                trigger_input = trigger_input[trigger_label == self.target_class]
                if len(trigger_input) == 0:
                    continue
            batch_conf = self.model.get_prob(trigger_input)[:, self.target_class].mean()
            confidence.update(batch_conf, len(trigger_input))
        return confidence.global_avg

    @torch.no_grad()
    def get_neuron_jaccard(self, k: int = None, ratio: float = 0.5) -> float:
        r"""Get Jaccard Index of neuron activations for feature maps
        between normal inputs and poison inputs.

        Find average top-k neuron indices of 2 kinds of feature maps
        ``clean_idx and poison_idx``, and return
        :math:`\frac{\text{len(clean\_idx \& poison\_idx)}}{\text{len(clean\_idx | poison\_idx)}}`

        Args:
            k (int): Top-k neurons to calculate jaccard index.
                Defaults to ``None``.
            ratio (float): Percentage of neurons if :attr:`k` is not provided.
                Defaults to ``0.5``.

        Returns:
            float: Jaccard Index.
        """
        clean_feats_list = []
        poison_feats_list = []
        for data in self.dataset.loader['valid']:
            _input, _label = self.model.get_data(data)
            trigger_input = self.add_mark(_input)

            clean_feats = self.model.get_fm(_input)
            poison_feats = self.model.get_fm(trigger_input)
            if clean_feats.dim() > 2:
                clean_feats = clean_feats.flatten(2).mean(2)
                poison_feats = poison_feats.flatten(2).mean(2)
            clean_feats_list.append(clean_feats)
            poison_feats_list.append(poison_feats)
        clean_feats_list = torch.cat(clean_feats_list).mean(dim=0)
        poison_feats_list = torch.cat(poison_feats_list).mean(dim=0)

        k = k or int(len(clean_feats_list) * ratio)
        clean_idx = set(clean_feats_list.argsort(
            descending=True)[:k].detach().cpu().tolist())
        poison_idx = set(poison_feats_list.argsort(
            descending=True)[:k].detach().cpu().tolist())
        jaccard_idx = len(clean_idx & poison_idx) / len(clean_idx | poison_idx)
        return jaccard_idx


class CleanLabelBackdoor(BackdoorAttack):
    r"""Backdoor attack abstract class of clean label.
    It inherits :class:`trojanvision.attacks.BackdoorAttack`.

    Under clean-label setting, only clean inputs from target class are infected,
    while the distortion is negligible for human to detect.
    """
    name = 'clean_label'

    def __init__(self, *args, train_mode: str = 'dataset', **kwargs):
        # monkey patch: to avoid calling get_poison_dataset() in super().__init__
        train_mode = 'batch'
        super().__init__(*args, train_mode=train_mode, **kwargs)
        self.target_set = self.dataset.get_dataset('train', class_list=[self.target_class])
        self.poison_num = int(self.poison_ratio * len(self.target_set))
        self.train_mode = 'dataset'

    def get_poison_dataset(self, poison_num: int = None, load_mark: bool = True,
                           seed: int = None) -> torch.utils.data.Dataset:
        r"""Get poison dataset from target class (no clean data).

        Args:
            poison_num (int): Number of poison data.
                Defaults to ``self.poison_num``
            load_mark (bool): Whether to load previously saved watermark.
                This should be ``False`` during attack.
                Defaults to ``True``.
            seed (int): Random seed to sample poison input indices.
                Defaults to ``env['data_seed']``.

        Returns:
            torch.utils.data.Dataset:
                Poison dataset from target class (no clean data).
        """
        file_path = os.path.join(self.folder_path, self.get_filename() + '.npy')
        if load_mark:
            if os.path.isfile(file_path):
                self.load_mark = False
                self.mark.load_mark(file_path, already_processed=True)
            else:
                raise FileNotFoundError(file_path)
        if seed is None:
            seed = env['data_seed']
        torch.random.manual_seed(seed)
        poison_num = min(poison_num or self.poison_num, len(self.target_set))
        _input, _label = sample_batch(self.target_set, batch_size=poison_num)
        _label = _label.tolist()
        trigger_input = self.add_mark(_input)
        return TensorListDataset(trigger_input, _label)


class DynamicBackdoor(BackdoorAttack):
    name = 'dynamic_backdoor'



#!/usr/bin/env python3

from trojanvision.models.imagemodel import _ImageModel, ImageModel
from trojanvision.datasets import ImageSet
import trojanvision.utils.model_archs.resnet_ap as resnet_ap

import torch
import torch.nn as nn
import torchvision.models
from torchvision.models.resnet import model_urls as urls
from torchvision.models.resnet import conv3x3
from collections import OrderedDict


class _ResNet(_ImageModel):

    def __init__(self, name: str = 'resnet18', dataset: ImageSet = None,
                 data_shape: list[int] = None, **kwargs):
        super().__init__(name=name, dataset=dataset, data_shape=data_shape, **kwargs)
        if data_shape is None:
            assert isinstance(dataset, ImageSet), 'Please specify data_shape or dataset'
            data_shape = dataset.data_shape
        module_list: list[nn.Module] = []
        if 's' in name.split('_'):
            from trojanvision.utils.model_archs.resnet_s import resnet_s
            _model = resnet_s(nclasses=self.num_classes)
            module_list.append(('conv1', _model.conv1))
            module_list.append(('bn1', _model.bn1))
            module_list.append(('relu', nn.ReLU(inplace=True)))
            self.classifier = nn.Sequential(OrderedDict([
                ('fc', _model.fc)
            ]))
        else:
            model_class = name.replace('_ap', '').replace('_comp', '').replace('_s', '')
            module = resnet_ap if 'ap' in name else torchvision.models
            ModelClass = getattr(module, model_class)
            kwargs = {'pool_size': data_shape[1] // 8} if 'ap' in name else {}
            _model: resnet_ap.ResNetAP | torchvision.models.ResNet = ModelClass(
                num_classes=self.num_classes, **kwargs)
            if '_comp' in name:
                conv1: nn.Conv2d = _model.conv1
                _model.conv1 = conv3x3(conv1.in_channels, conv1.out_channels)
                if 'resnext' in name:
                    _model.fc = nn.Linear(_model.fc.in_features // 2,
                                          _model.fc.out_features,
                                          bias=_model.fc.bias is None)
                module_list.append(('conv1', _model.conv1))
                module_list.append(('bn1', _model.bn1))
                module_list.append(('relu', _model.relu))
            else:
                module_list.append(('conv1', _model.conv1))
                module_list.append(('bn1', _model.bn1))
                module_list.append(('relu', _model.relu))
                module_list.append(('maxpool', _model.maxpool))
            self.pool = _model.avgpool  # nn.AdaptiveAvgPool2d((1, 1))
            self.classifier = nn.Sequential(OrderedDict([
                ('fc', _model.fc)  # nn.Linear(512 * block.expansion, num_classes)
            ]))
            # block.expansion = 1 if BasicBlock and 4 if Bottleneck
            # ResNet 18,34 use BasicBlock, 50 and higher use Bottleneck
        module_list.extend([('layer1', _model.layer1),
                            ('layer2', _model.layer2),
                            ('layer3', _model.layer3)])
        if not ('comp' in name and 'resnext' in name):
            module_list.append(('layer4', _model.layer4))
        self.features = nn.Sequential(OrderedDict(module_list))


class ResNet(ImageModel):
    r"""ResNet model series including ResNet, ResNext and WideResNet.

    :Available model names:

        .. code-block:: python3

            ['resnet', 'resnet_comp', 'resnet_s',
             'resnet18', 'resnet34', 'resnet50', 'resnet101', 'resnet152',
             'resnet18_comp', 'resnet34_comp', 'resnet50_comp', 'resnet101_comp', 'resnet152_comp',
             'resnext50_32x4d', 'resnext101_32x8d',
             'resnext50_32x4d_comp', 'resnext101_32x8d_comp',
             'wide_resnet50_2', 'wide_resnet101_2',
             'wide_resnet50_2_comp', 'wide_resnet101_2_comp',

             'resnet18_s', 'resnet34_s', 'resnet50_s', 'resnet101_s', 'resnet152_s',
             'resnet18_ap_comp']

    See Also:
        * ResNet:

          - torchvision: :any:`torchvision.models.resnet18`
          - paper: `Deep Residual Learning for Image Recognition`_
        * ResNext:

          - torchvision: :any:`torchvision.models.resnext50_32x4d`
          - paper: `Aggregated Residual Transformations for Deep Neural Networks`_
        * WideResNet:

          - torchvision: :any:`torchvision.models.wide_resnet50_2`
          - paper: `Wide Residual Networks`_

    Note:
        * ``_comp`` reduces the first convolutional layer
          from ``kernel_size=7, stride=2, padding=3``

          to ``kernel_size=3, stride=1, padding=1``,
          and removes the ``maxpool`` layer before block layers.
        * ``_s`` further reduces the conv channels and number of blocks based on ``_comp``.

          ``ResNetS`` is used in NIPS 2017 paper by Facebook Research about continual learning.

          - paper: `Gradient Episodic Memory for Continual Learning`_ (GEM)
          - code: https://github.com/facebookresearch/GradientEpisodicMemory/blob/master/model/common.py
        * ``_ap`` conducts average pooling in all blocks
          rather than a global pooling layer after feature extractor.

          ``ResNetAP`` is used in ICLR 2021 paper about dataset condensation.

          - paper: `Dataset Condensation with Gradient Matching`_
          - code: https://github.com/VICO-UoE/DatasetCondensation/blob/master/networks.py

    .. _Deep Residual Learning for Image Recognition:
        https://arxiv.org/abs/1512.03385
    .. _Aggregated Residual Transformations for Deep Neural Networks:
        https://arxiv.org/abs/1611.05431
    .. _Wide Residual Networks:
        https://arxiv.org/abs/1605.07146
    .. _Gradient Episodic Memory for Continual Learning:
        https://arxiv.org/abs/1706.08840
    .. _Dataset Condensation with Gradient Matching:
        https://arxiv.org/abs/2006.05929
    """
    available_models = ['resnet', 'resnet_comp', 'resnet_s',
                        'resnet18', 'resnet34', 'resnet50', 'resnet101', 'resnet152',
                        'resnet18_comp', 'resnet34_comp', 'resnet50_comp', 'resnet101_comp', 'resnet152_comp',
                        'resnext50_32x4d', 'resnext101_32x8d',
                        'resnext50_32x4d_comp', 'resnext101_32x8d_comp',
                        'wide_resnet50_2', 'wide_resnet101_2',
                        'wide_resnet50_2_comp', 'wide_resnet101_2_comp',

                        'resnet18_s', 'resnet34_s', 'resnet50_s', 'resnet101_s', 'resnet152_s',
                        'resnet18_ap_comp']
    model_urls = urls

    def __init__(self, name: str = 'resnet', layer: int = 18,
                 model: type[_ResNet] = _ResNet,
                 **kwargs):
        super().__init__(name=name, layer=layer, model=model, **kwargs)

    def get_official_weights(self, **kwargs) -> OrderedDict[str, torch.Tensor]:
        _dict = super().get_official_weights(**kwargs)
        new_dict = OrderedDict()
        for i, (key, value) in enumerate(_dict.items()):
            prefix = 'features.' if i < len(_dict) - 2 else 'classifier.'
            new_dict[prefix + key] = value
        return new_dict

    @classmethod
    def get_name(cls, name: str, layer: int = '') -> str:
        prefix = ''
        if name.startswith('wide_'):
            prefix = 'wide_'
            name = name[5:]
        return prefix + super().get_name(name, layer=layer)



#!/usr/bin/env python3

from trojanzoo.utils.output import prints

from typing import Generic, MutableMapping, TypeVar
_KT = TypeVar("_KT")  # Key type.
_VT = TypeVar("_VT")  # Value type.


# TODO: issue 3 why need Generic
class Module(MutableMapping[_KT, _VT], Generic[_KT, _VT]):
    r"""A dict-like class which supports attribute-like view as well.

    Args:
        *args: Positional dict-like arguments.
            All keys will be merged together.
        **kwargs: Keyword arguments that compose a dict.
            All keys will be merged together.

    Attributes:
        _marker (str): The marker of the class,
            which is shown in ``str(self)``.
            Defaults to ``'M'``.
    """
    _marker: str = 'M'

    def __init__(self, *args: MutableMapping[_KT, _VT], **kwargs: _VT):
        self.__data: dict[_KT, _VT] = {}
        if len(args) == 1 and args[0] is None:
            return
        self.update(*args, **kwargs)

    def update(self, *args: MutableMapping[_KT, _VT], **kwargs: _VT):
        r"""update values.

        Args:
            *args: Positional dict-like arguments.
                All keys will be merged together.
            **kwargs: Keyword arguments that compose a dict.
                All keys will be merged together.

        Returns:
            Module: return :attr:`self` for stream usage.
        """
        args: list = list(args)     # TODO: issue 2 pylance issue
        args.append(kwargs)
        for module in args:
            self._update(module)
        return self

    # TODO: issue 4 dict | Module
    def _update(self, module: MutableMapping[_KT, _VT]):
        for key, value in module.items():
            if value is None:
                continue
            if key in self.keys() and isinstance(self[key], Module):
                sub_module: Module = self[key]
                sub_module.update(value)
            elif key in self.keys() and isinstance(value, Module):
                self[key] = type(value)(self[key]).update(value)
            elif isinstance(value, Module):
                self[key] = value.copy()
            else:
                self[key] = value
        return self

    def remove_none(self):
        r"""Remove the parameters whose values are ``None``.

        Returns:
            Module: return :attr:`self` for stream usage.
        """
        for key in self.__data.keys():
            if self.__data[key] is None:
                del self.__data[key]
        return self

    def copy(self):
        r"""Deepcopy of :attr:`self`.

        Returns:
            Module: return the deepcopy of :attr:`self`.
        """
        return type(self)(self)

    def clear(self):
        r"""Remove all keys.

        Returns:
            Module: return :attr:`self` for stream usage.
        """
        self.__data = {}
        return self

    def keys(self):
        return self.__data.keys()

    def items(self):
        return self.__data.items()

    def __getattr__(self, name: _KT) -> _VT:
        if '__data' in name:
            return super().__getattr__(name)
        return self.__data[name]

    def __getitem__(self, k: _KT):
        return self.__data[k]

    def __setattr__(self, name: _KT, value: _VT):
        if '__data' in name:
            return super().__setattr__(name, value)
        self.__data[name] = value

    def __setitem__(self, k: _KT, v: _VT):
        self.__data[k] = v

    def __delattr__(self, name: _KT):
        del self.__data[name]

    def __delitem__(self, v: _KT):
        del self.__data[v]

    def __str__(self):
        return self._marker + self.__data.__str__()

    def __repr__(self):
        return self._marker + self.__data.__repr__()

    def __len__(self):
        return self.__data.__len__()

    def __iter__(self):
        return self.__data.__iter__()

    def summary(self, indent: int = 0):
        r"""Output information of :attr:`self`.

        Args:
            indent (int): The space indent for the entire string.
                Defaults to ``0``.
        """
        prints(self, indent=indent)


# TODO: issue 3 why need Generic, Module[_KT, _VT]
class Param(Module, Generic[_KT, _VT]):
    r"""A dict-like class to store parameters config that
    inherits :class:`Module` and further extends default values.
    You can view and set keys by attributes as well.

    Args:
        *args: Positional dict-like arguments.
            All keys will be merged together.
            If there is only 1 argument and no keyword argument,
            regard it as the default value.
        **kwargs: Keyword arguments that compose a dict.
            All keys will be merged together.

    Attributes:
        _marker (str): The marker of the class,
            which is shown in ``str(self)``.
            Defaults to ``'M'``.
        default (Any): The default value of unknown keys.
    """
    _marker = 'P'

    def update(self, *args: dict[_KT, _VT], **kwargs: _VT):
        if len(kwargs) == 0 and len(args) == 1 and \
                not isinstance(args[0], (dict, Module)):
            self.default = args[0]
            return self
        return super().update(*args, **kwargs)

    def _update(self, module: dict[_KT, _VT]):
        for key, value in module.items():
            if key == 'default':
                self.default = value
        super()._update(module)
        return self     # For linting purpose

    def remove_none(self):
        for key in list(self.__data.keys()):
            if self.__data[key] is None and \
                    not (isinstance(key, str) and key == 'default'):
                del self.__data[key]
        return self

    def __getattr__(self, name: str) -> _VT:
        try:
            return super().__getattr__(name)
        except KeyError:
            if 'default' in self.keys():
                return self['default']
            raise

    def __getitem__(self, key: str) -> _VT:
        if key not in self.keys():
            key = 'default'
            if 'default' not in self.keys():
                print(self)
                raise KeyError(key)
        return super().__getitem__(key)

    def clear(self):
        super().clear()
        self.default = None
        return self



#!/usr/bin/env python3

import torch
import torch.nn as nn
import torch.nn.utils.parametrize as P


class Std(nn.Module):
    def forward(self, X: torch.Tensor):
        v, m = torch.var_mean(X, dim=[1, 2, 3], keepdim=True, unbiased=False)
        return (X - m) / torch.sqrt(v + 1e-10)


class StdConv2d(nn.Conv2d):
    def __init__(self, *args, parametrize: bool = True, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.parametrize = parametrize
        if parametrize:
            P.register_parametrization(self, 'weight', Std())

    def parametrize_(self, parametrize: bool = True):
        if parametrize:
            if not self.parametrize:
                P.register_parametrization(self, 'weight', Std())
        elif self.parametrize:
            P.remove_parametrizations(self, 'weight')
        self.parametrize = parametrize
        return self



#!/usr/bin/env python3

"""
CUDA_VISIBLE_DEVICES=0 python examples/train.py --color --verbose 1 --dataset cifar10 --model darts --supernet --arch_search --arch_unrolled --layers 8 --init_channels 16 --batch_size 64 --lr 0.025 --lr_scheduler --lr_min 1e-3 --grad_clip 5.0 --epochs 50
"""  # noqa: E501

import trojanvision.utils.model_archs.darts as darts
from trojanvision.datasets import ImageSet
from trojanvision.models.imagemodel import _ImageModel, ImageModel
from trojanvision.utils.model_archs.darts import FeatureExtractor, AuxiliaryHead, Genotype, genotypes
from trojanvision.utils.model_archs.darts.operations import PRIMITIVES

import torch
import torch.nn as nn
from torchvision.datasets.utils import download_file_from_google_drive
import os
import itertools
from collections import OrderedDict

from typing import TYPE_CHECKING
from trojanzoo.utils.fim import KFAC, EKFAC
from trojanzoo.utils.model import ExponentialMovingAverage
from torch.optim import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
import torch.utils.data
import argparse  # TODO: python 3.10
from collections.abc import Callable
if TYPE_CHECKING:
    pass

url = {
    'cifar10': '1Y13i4zKGKgjtWBdC0HWLavjO7wvEiGOc',
    'ptb': '1Mt_o6fZOlG-VDF3Q5ModgnAJ9W6f_av2',
    'imagenet': '1AKr6Y_PoYj7j0Upggyzc26W0RVdg4CVX'
}


def _concat(xs: torch.Tensor) -> torch.Tensor:
    return torch.cat([x.flatten() for x in xs])


class _DARTS(_ImageModel):
    def __init__(self, auxiliary: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.features: FeatureExtractor | darts.search.FeatureExtractor
        self.auxiliary_head: nn.Sequential = None
        if auxiliary:
            self.auxiliary_head = AuxiliaryHead(C=self.features.aux_dim, num_classes=self.num_classes)
        num_features = kwargs.get('num_features', [self.features.feats_dim])
        self.classifier = self.define_classifier(num_features=num_features, num_classes=self.num_classes)

    @staticmethod
    def define_features(supernet: bool = False,
                        genotype: Genotype = genotypes.darts,
                        init_channels: int = 36, layers: int = 20,
                        dropout_p: float = 0.2, **kwargs
                        ) -> FeatureExtractor | darts.search.FeatureExtractor:
        if supernet:
            return darts.search.FeatureExtractor(init_channels, layers, **kwargs)
        return FeatureExtractor(genotype, init_channels, layers, dropout_p, **kwargs)


class DARTS(ImageModel):
    r"""DARTS-like models used in Neural Architecture Search.

    :Available model names:

        .. code-block:: python3

            ['darts']

    See Also:
        * paper: `DARTS\: Differentiable Architecture Search`_
        * code: https://github.com/quark0/darts

    Args:
        supernet (bool): Whether to use supernet (mixed operations).
            Defaults to ``False``.
        model_arch (str): Genotype name in ``trojanvision.utils.model_archs.genotypes`` to use.
            Defaults to ``'darts'``.

            * ``'amoebanet', 'amoebanet_adapt'``
            * ``'darts_v1', 'darts_v2'('darts')``
            * ``'drnas_cifar10'('drnas'), 'drnas_imagenet'``
            * ``'enas', 'enas_adapt'``
            * ``'nasnet', 'nasnet_adapt'``
            * ``'pc_darts_cifar'('pc_darts'), 'pc_darts_image'``
            * ``'pdarts'``
            * ``'robust_darts'``
            * ``'sgas'``
            * ``'snas_mild', 'snas_adapt'``
            * ``'random'``
            * ``'diy_deep', 'diy_noskip', 'diy_deep_noskip'``
        layers (int): Total number of layers. Defaults to ``20``.
        init_channels (int): :attr:`out_channel` of stem conv layer.
            Defaults to ``36``.
        dropout_p (float): Dropout probability.
            Defaults to ``0.2``.
        auxiliary (bool): Whether to use auxiliary classifier.
            Defaults to ``False``.
        auxiliary_weight (float): Loss weight of auxiliary classifier.
            Defaults to ``0.4``.
        arch_search (bool): Whether to search supernet architecture weight parameters.
            Defaults to ``False``.
        use_full_train_set (bool): Whether to use full training data during architecture search.
            Defaults to ``False``.
        arch_lr (float): Learning rate for architecture optimizer.
            Defaults to ``3e-4``
        arch_weight_decay (float): Weight decay for architecture optimizer.
            Defaults to ``1e-3``.
        arch_unrolled (bool): Whether to use one-step unrolled validation loss (darts-v2).
            Defaults to ``False``.

    Attributes:
        genotype (Genotype): Genotype of cell architecture.

    Note:
        The implementation of DARTS model is in ``trojanvision.utils.model_archs.darts``

    .. _DARTS\: Differentiable Architecture Search:
        https://arxiv.org/abs/1806.09055
    """
    available_models = ['darts']

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--supernet', action='store_true', help='whether to use supernet')
        group.add_argument('--model_arch', help='genotype name (default: "darts")')
        group.add_argument('--layers', type=int, help='total number of layers (default: 20)')
        group.add_argument('--init_channels', type=int, help='out_channel of stem conv layer (default: 36)')
        group.add_argument('--dropout_p', type=float, help='dropout probability (default: 0.2)')
        group.add_argument('--auxiliary', action='store_true', help='whether to use auxiliary classifier')
        group.add_argument('--auxiliary_weight', type=float,
                           help='loss weight of auxiliary classifier (default: 0.4)')
        group.add_argument('--arch_search', action='store_true',
                           help='whether to search supernet architecture weight parameters')
        group.add_argument('--use_full_train_set', action='store_true',
                           help='whether to use full training data during architecture search')
        group.add_argument('--arch_lr', type=float,
                           help='learning rate for architecture optimizer (default: 3e-4)')
        group.add_argument('--arch_weight_decay', type=float,
                           help='weight decay for architecture optimizer (default: 1e-3)')
        group.add_argument('--arch_unrolled', action='store_true', default=False,
                           help='whether to use one-step unrolled validation loss (darts-v2)')
        return group

    def __init__(self, name: str = 'darts', model_arch: str = 'darts',
                 layers: int = 20, init_channels: int = 36, dropout_p: float = 0.2,
                 auxiliary: bool = False, auxiliary_weight: float = 0.4,
                 genotype: Genotype = None, model: type[_DARTS] = _DARTS,
                 supernet: bool = False, arch_search: bool = False,
                 use_full_train_set: bool = False,
                 arch_lr: float = 3e-4, arch_weight_decay=1e-3,
                 arch_unrolled: bool = False,
                 primitives=PRIMITIVES, **kwargs):
        # TODO: ImageNet parameter settings
        self.supernet = supernet
        self.arch_search = arch_search
        if supernet:
            name += '_supernet'
        if not supernet and genotype is None:
            model_arch = model_arch.lower()
            name = model_arch
            try:
                genotype = getattr(genotypes, model_arch)
            except AttributeError:
                print('Available Model Architectures: ')
                model_arch_list = [element for element in dir(genotypes)
                                   if '__' not in element and
                                   element not in ['Genotype', 'namedtuple']]
                print(model_arch_list)
                raise
        self.layers = layers
        self.init_channels = init_channels
        self.dropout_p = dropout_p
        self.auxiliary = auxiliary
        self.auxiliary_weight = auxiliary_weight
        self.arch_lr = arch_lr
        self.arch_weight_decay = arch_weight_decay
        self.arch_unrolled = arch_unrolled
        super().__init__(name=name, layers=layers, init_channels=init_channels, dropout_p=dropout_p,
                         genotype=genotype, model=model,
                         auxiliary=auxiliary, primitives=primitives,
                         supernet=supernet, **kwargs)
        self._model: _DARTS
        self.param_list['darts'] = ['supernet', 'layers', 'init_channels']
        if not supernet:
            self.param_list['darts'].extend(['dropout_p', 'genotype'])
        if auxiliary:
            self.param_list['darts'].insert(1, 'auxiliary_weight')

        if supernet and arch_search:
            self.use_full_train_set = use_full_train_set
            for alpha in self.arch_parameters():
                alpha.requires_grad_()
            train2, train3 = self.dataset.split_dataset(
                self.dataset.get_dataset('train'),
                percent=0.5)
            self.train2 = self.dataset.get_dataloader(
                mode='train', dataset=train2)
            self.train3 = self.dataset.get_dataloader(
                mode='train', dataset=train3)
            self.valid_iterator = itertools.cycle(self.train3)
            self.arch_optimizer = torch.optim.Adam(self.arch_parameters(),
                                                   lr=arch_lr, betas=(0.5, 0.999),
                                                   weight_decay=arch_weight_decay)
            self.param_list['arch_search'] = ['use_full_train_set', 'arch_optimizer']

    @property
    def genotype(self) -> Genotype:
        return self._model.features.genotype() if self.supernet else self._model.features.genotype

    def loss(self, _input: torch.Tensor = None, _label: torch.Tensor = None,
             _output: torch.Tensor = None, amp: bool = False, **kwargs) -> torch.Tensor:
        if self.auxiliary:
            assert isinstance(self._model.auxiliary_head, nn.Sequential)
            if amp:
                with torch.cuda.amp.autocast():
                    return self.loss_with_aux(_input, _label)
            return self.loss_with_aux(_input, _label)
        return super().loss(_input, _label, _output, amp=amp, **kwargs)

    def loss_with_aux(self, _input: torch.Tensor = None, _label: torch.Tensor = None) -> torch.Tensor:
        feats, feats_aux = self._model.features.forward_with_aux(self._model.preprocess(_input))
        logits: torch.Tensor = self._model.classifier(self._model.flatten(self._model.pool(feats)))
        logits_aux: torch.Tensor = self._model.auxiliary_head(feats_aux)
        return super().loss(_output=logits, _label=_label) \
            + self.auxiliary_weight * super().loss(_output=logits_aux, _label=_label)

    def load(self, *args, strict: bool = False, **kwargs):
        return super().load(*args, strict=strict, **kwargs)

    def get_official_weights(self, dataset: str = None, **kwargs) -> OrderedDict[str, torch.Tensor]:
        assert str(self.genotype) == str(genotypes.darts)
        if dataset is None and isinstance(self.dataset, ImageSet):
            dataset = self.dataset.name
        file_name = f'darts_{dataset}.pt'
        folder_path = os.path.join(torch.hub.get_dir(), 'darts')
        file_path = os.path.join(folder_path, file_name)
        download_file_from_google_drive(file_id=url[dataset], root=folder_path, filename=file_name)
        print('get official model weights from Google Drive: ', url[dataset])
        _dict: OrderedDict[str, torch.Tensor] = torch.load(file_path,
                                                           map_location='cpu')
        if 'state_dict' in _dict.keys():
            _dict = _dict['state_dict']

        new_dict: OrderedDict[str, torch.Tensor] = self.state_dict()
        old_keys = list(_dict.keys())
        new_keys = list(new_dict.keys())
        new2old: dict[str, str] = {}
        i = 0
        j = 0
        while(i < len(new_keys) and j < len(old_keys)):
            if 'num_batches_tracked' in new_keys[i]:
                i += 1
                continue
            if 'auxiliary_head' not in new_keys[i] and 'auxiliary_head' in old_keys[j]:
                j += 1
                continue
            new2old[new_keys[i]] = old_keys[j]
            i += 1
            j += 1
        for i, key in enumerate(new_keys):
            if 'num_batches_tracked' in key:
                new_dict[key] = torch.tensor(0)
            else:
                new_dict[key] = _dict[new2old[key]]
        return new_dict

    def arch_parameters(self) -> list[torch.Tensor]:
        return self._model.features.arch_parameters()

    def named_arch_parameters(self) -> list[tuple[str, torch.Tensor]]:
        return self._model.features.named_arch_parameters()

    def _train(self, epochs: int, optimizer: Optimizer, lr_scheduler: _LRScheduler = None,
               adv_train: bool = None,
               lr_warmup_epochs: int = 0,
               model_ema: ExponentialMovingAverage = None,
               model_ema_steps: int = 32,
               grad_clip: float = None, pre_conditioner: None | KFAC | EKFAC = None,
               print_prefix: str = 'Epoch', start_epoch: int = 0, resume: int = 0,
               validate_interval: int = 10, save: bool = False, amp: bool = False,
               loader_train: torch.utils.data.DataLoader = None,
               loader_valid: torch.utils.data.DataLoader = None,
               epoch_fn: Callable[..., None] = None,
               get_data_fn: Callable[...,
                                     tuple[torch.Tensor, torch.Tensor]] = None,
               loss_fn: Callable[..., torch.Tensor] = None,
               after_loss_fn: Callable[..., None] = None,
               validate_fn: Callable[..., tuple[float, float]] = None,
               save_fn: Callable[..., None] = None, file_path: str = None,
               folder_path: str = None, suffix: str = None,
               writer=None, main_tag: str = 'train', tag: str = '',
               accuracy_fn: Callable[..., list[float]] = None,
               verbose: bool = True, indent: int = 0, **kwargs):
        get_data_fn = get_data_fn or self.get_data
        validate_fn = validate_fn or self._validate
        if self.arch_search:
            if not self.use_full_train_set:
                loader_train = loader_train or self.train2
            if self.arch_unrolled:
                self.optimizer = optimizer
                # self.lr_scheduler = lr_scheduler

            get_data_old = get_data_fn
            validate_old = validate_fn

            def get_data(data: tuple[torch.Tensor, torch.Tensor], adv_train: bool = False,
                         mode: str = 'train', **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
                _input, _label = get_data_old(data, adv_train=adv_train, **kwargs)
                if mode == 'train':
                    data_valid = next(self.valid_iterator)
                    input_valid, label_valid = get_data_old(data_valid, adv_train=adv_train, **kwargs)
                    self.arch_optimizer.zero_grad()
                    if self.arch_unrolled:
                        self._backward_step_unrolled(_input, _label, input_valid, label_valid)
                    else:
                        loss = self.loss(input_valid, label_valid)
                        loss.backward(inputs=self.arch_parameters())
                    self.arch_optimizer.step()
                return _input, _label

            def _validate(adv_train: bool = None,
                          loader: torch.utils.data.DataLoader = None,
                          **kwargs) -> tuple[float, float]:
                print(self.genotype)
                # if not self.use_full_train_set:
                #     super()._validate(loader=self.train3,
                #                       adv_train=adv_train,
                #                       print_prefix='TrainVal', **kwargs)
                return validate_old(loader=loader, adv_train=adv_train, **kwargs)

            get_data_fn = get_data
            validate_fn = _validate

        return super()._train(epochs=epochs, optimizer=optimizer, lr_scheduler=lr_scheduler,
                              adv_train=adv_train,
                              lr_warmup_epochs=lr_warmup_epochs,
                              model_ema=model_ema, model_ema_steps=model_ema_steps,
                              grad_clip=grad_clip, pre_conditioner=pre_conditioner,
                              print_prefix=print_prefix, start_epoch=start_epoch,
                              resume=resume, validate_interval=validate_interval,
                              save=save, amp=amp,
                              loader_train=loader_train, loader_valid=loader_valid,
                              epoch_fn=epoch_fn, get_data_fn=get_data_fn,
                              loss_fn=loss_fn, after_loss_fn=after_loss_fn,
                              validate_fn=validate_fn,
                              save_fn=save_fn, file_path=file_path,
                              folder_path=folder_path, suffix=suffix,
                              writer=writer, main_tag=main_tag, tag=tag,
                              accuracy_fn=accuracy_fn,
                              verbose=verbose, indent=indent, **kwargs)

    def _backward_step_unrolled(self, input_train: torch.Tensor, target_train: torch.Tensor,
                                input_valid: torch.Tensor, target_valid: torch.Tensor,
                                optimizer: Optimizer = None):
        optimizer = optimizer or self.optimizer
        eta = optimizer.param_groups[0]['lr']
        # if self.lr_scheduler is not None:
        #     eta = self.lr_scheduler.get_last_lr()
        w = self.state_dict()
        self._compute_unrolled_model(input_train, target_train, eta,
                                     optimizer=optimizer)
        unrolled_loss = self.loss(input_valid, target_valid)
        unrolled_loss.backward()
        dalpha = [v.grad for v in self.arch_parameters()]
        vector = [v.grad.data for v in self.parameters()]

        self.load_state_dict(w)
        implicit_grads = self._hessian_vector_product(vector, input_train, target_train)

        for g, ig in zip(dalpha, implicit_grads):
            g.data.sub_(ig, alpha=eta)

        for v, g in zip(self.arch_parameters(), dalpha):
            if v.grad is None:
                v.grad = g.data
            else:
                v.grad.copy_(g)

    # def _construct_model_from_theta(self, theta):
    #     model_new = self.model.new()
    #     model_dict = self.model.state_dict()

    #     params, offset = {}, 0
    #     for k, v in self.model.named_parameters():
    #         v_length = np.prod(v.size())
    #         params[k] = theta[offset: offset + v_length].view(v.size())
    #         offset += v_length

    #     assert offset == len(theta)
    #     model_dict.update(params)
    #     model_new.load_state_dict(model_dict)
    #     return model_new.cuda()

    def _compute_unrolled_model(self, input: torch.Tensor, target: torch.Tensor,
                                eta: float, optimizer: Optimizer = None):
        optimizer = optimizer or self.optimizer
        weight_decay = optimizer.param_groups[0]['weight_decay']
        loss = self.loss(input, target)
        gtheta = torch.autograd.grad(loss, self.parameters())
        with torch.no_grad():
            dtheta = [grad + weight_decay * param for grad, param in zip(gtheta, self.parameters())]
            try:
                momentum = optimizer.param_groups[0]['momentum']
                dtheta = [dtheta_i + momentum * optimizer.state[v]['momentum_buffer']
                          for dtheta_i, v in zip(dtheta, self.model.parameters())]
            except KeyError:
                pass
            for param, delta in zip(self.parameters(), dtheta):
                param.data.sub_(delta, alpha=eta)

    def _hessian_vector_product(self, vector: list[torch.Tensor],
                                _input: torch.Tensor, target: torch.Tensor,
                                r: float = 1e-2
                                ) -> list[torch.Tensor]:
        arch_params = self.arch_parameters()
        R = r / _concat(vector).norm()
        for p, v in zip(self.parameters(), vector):
            p.data.add_(v, alpha=R)
        loss = self.loss(_input, target)
        grads_p = list(torch.autograd.grad(loss, arch_params, allow_unused=True))

        for p, v in zip(self.parameters(), vector):
            p.data.sub_(v, alpha=2 * R)
        loss = self.loss(_input, target)
        grads_n = list(torch.autograd.grad(loss, arch_params, allow_unused=True))

        for p, v in zip(self.parameters(), vector):
            p.data.add_(v, alpha=R)

        return [(x - y) / (2 * R) for x, y in zip(grads_p, grads_n)]



#!/usr/bin/env python3

from ...abstract import InputFiltering
from trojanzoo.utils.logger import MetricLogger
from trojanzoo.utils.data import TensorListDataset

import torch
import argparse


class Strip(InputFiltering):
    name: str = 'strip'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--strip_fpr', type=float,
                           help='false positive rate for strip defense '
                           '(default: 0.05)')
        group.add_argument('--strip_alpha', type=float,
                           help='alpha value of each image '
                           'during blending for strip defense '
                           '(default: 0.5)')
        group.add_argument('--strip_sample_num', type=int,
                           help='number of sampled images to blend '
                           'with each test image for strip defense '
                           '(default: 64)')
        return group

    def __init__(self, strip_fpr: float = 0.05,
                 strip_alpha: float = 0.5,
                 strip_sample_num: int = 64, **kwargs):
        super().__init__(**kwargs)
        self.param_list['strip'] = ['strip_fpr', 'strip_alpha', 'strip_sample_num']
        self.strip_fpr = strip_fpr
        self.strip_alpha: float = strip_alpha
        self.strip_sample_num: int = strip_sample_num
        self.loader = self.dataset.get_dataloader(mode='valid', drop_last=True)

    @torch.no_grad()
    def get_pred_labels(self) -> torch.Tensor:
        r"""Get predicted labels for test inputs.

        Returns:
            torch.Tensor: ``torch.BoolTensor`` with shape ``(2 * defense_input_num)``.
        """
        logger = MetricLogger(meter_length=40)
        str_format = '{global_avg:5.3f} ({min:5.3f}, {max:5.3f})'
        logger.create_meters(clean_score=str_format, poison_score=str_format)
        test_set = TensorListDataset(self.test_input, self.test_label)
        test_loader = self.dataset.get_dataloader(mode='valid', dataset=test_set)
        for data in logger.log_every(test_loader):
            _input, _label = self.model.get_data(data)
            trigger_input = self.attack.add_mark(_input)
            logger.meters['clean_score'].update_list(self.get_score(_input).tolist())
            logger.meters['poison_score'].update_list(self.get_score(trigger_input).tolist())
        clean_score = torch.as_tensor(logger.meters['clean_score'].deque)
        poison_score = torch.as_tensor(logger.meters['poison_score'].deque)
        clean_score_sorted = clean_score.msort()
        threshold_low = float(clean_score_sorted[int(self.strip_fpr * len(poison_score))])
        entropy = torch.cat((clean_score, poison_score))
        print(f'Threshold: {threshold_low:5.3f}')
        return torch.where(entropy < threshold_low,
                           torch.ones_like(entropy).bool(),
                           torch.zeros_like(entropy).bool())

    def get_score(self, _input: torch.Tensor) -> torch.Tensor:
        _list = []
        for i, data in enumerate(self.loader):
            if i >= self.strip_sample_num:
                break
            benign_input, _ = self.model.get_data(data)
            benign_input = benign_input[:len(_input)]
            test_input = self.strip_alpha * (_input - benign_input) + benign_input
            test_output = self.model(test_input)
            test_entropy = -test_output.softmax(1).mul(test_output.log_softmax(1)).sum(1)
            _list.append(test_entropy.cpu())
        return torch.stack(_list).mean(0)



#!/usr/bin/env python3

from .imageset import ImageSet
from .imagefolder import ImageFolder

from .folder import *
from .normal import *

from . import folder, normal

from trojanvision.configs import config
import trojanzoo.datasets

import argparse
from trojanzoo.configs import Config

module_list = [folder, normal]
__all__ = ['ImageSet', 'ImageFolder', 'class_dict', 'add_argument', 'create']
class_dict: dict[str, type[ImageSet]] = {}
for module in module_list:
    __all__.extend(module.__all__)
    class_dict.update(module.class_dict)


def add_argument(parser: argparse.ArgumentParser,
                 dataset_name: str = None, dataset: str | ImageSet = None,
                 config: Config = config, class_dict: dict[str, type[ImageSet]] = class_dict
                 ) -> argparse._ArgumentGroup:
    r"""
    | Add image dataset arguments to argument parser.
    | For specific arguments implementation, see :meth:`ImageSet.add_argument()`.

    Args:
        parser (argparse.ArgumentParser): The parser to add arguments.
        dataset_name (str): The dataset name.
        dataset (str | Dataset): Dataset instance or dataset name
            (as the alias of `dataset_name`).
        config (Config): The default parameter config,
            which contains the default dataset name if not provided.
        class_dict (dict[str, type[Dataset]]):
            Map from dataset name to dataset class.
            Defaults to ``trojanvision.datasets.class_dict``.

    See Also:
        :func:`trojanzoo.datasets.add_argument()`
    """
    return trojanzoo.datasets.add_argument(parser=parser, dataset_name=dataset_name, dataset=dataset,
                                           config=config, class_dict=class_dict)


def create(dataset_name: str = None, dataset: str = None,
           config: Config = config, class_dict: dict[str, type[ImageSet]] = class_dict,
           **kwargs) -> ImageSet:
    r"""
    | Create a image dataset instance.
    | For arguments not included in :attr:`kwargs`,
      use the default values in :attr:`config`.
    | The default value of :attr:`folder_path` is
      ``'{data_dir}/{data_type}/{name}'``.
    | For dataset implementation, see :class:`ImageSet`.

    Args:
        dataset_name (str): The dataset name.
        dataset (str): The alias of `dataset_name`.
        config (Config): The default parameter config.
        class_dict (dict[str, type[ImageSet]]):
            Map from dataset name to dataset class.
            Defaults to ``trojanvision.datasets.class_dict``.
        **kwargs: Keyword arguments
            passed to dataset init method.

    Returns:
        ImageSet: Image dataset instance.

    See Also:
        :func:`trojanzoo.datasets.create()`
    """
    return trojanzoo.datasets.create(dataset_name=dataset_name, dataset=dataset,
                                     config=config, class_dict=class_dict, **kwargs)



#!/usr/bin/env python3

from trojanvision.datasets.imagefolder import ImageFolder
from trojanzoo.utils.module import Module

from torchvision import datasets
import os
import json

from trojanvision import __file__ as root_file
root_dir = os.path.dirname(root_file)


class ImageNet(ImageFolder):
    r"""ImageNet (ILSVRC2012) dataset introduced by Jia Deng and Feifei Li in 2012.
    It inherits :class:`trojanvision.datasets.ImageFolder`.

    See Also:
        * torchvision: :any:`torchvision.datasets.ImageNet`
        * paper: `ImageNet\: A large-scale hierarchical image database`_
        * website: https://image-net.org/about.php

    Note:
        According to https://github.com/pytorch/vision/issues/1563,
        You need to personally visit https://image-net.org/download-images.php
        to download the dataset.

        Expected files:

            * ``'{self.folder_path}/ILSVRC2012_devkit_t12.tar.gz'``
            * ``'{self.folder_path}/ILSVRC2012_img_train.tar'``
            * ``'{self.folder_path}/ILSVRC2012_img_val.tar'``
            * ``'{self.folder_path}/meta.bin'``

    Attributes:
        name (str): ``'imagenet'``
        num_classes (int): ``1000``
        data_shape (list[int]): ``[3, 224, 224]``
        norm_par (dict[str, list[float]]):
            | ``{'mean': [0.485, 0.456, 0.406],``
            | ``'std'  : [0.229, 0.224, 0.225]}``

    .. _ImageNet\: A large-scale hierarchical image database:
        https://ieeexplore.ieee.org/document/5206848
    """

    name = 'imagenet'
    url = {
        'train': 'http://www.image-net.org/challenges/LSVRC/2012/nnoupb/ILSVRC2012_img_train.tar',
        'valid': 'http://www.image-net.org/challenges/LSVRC/2012/nnoupb/ILSVRC2012_img_val.tar',
        'test': 'http://www.image-net.org/challenges/LSVRC/2012/nnoupb/ILSVRC2012_img_test.tar',
    }
    md5 = {
        'train': '1d675b47d978889d74fa0da5fadfb00e',
        'valid': '29b22e2961454d5413ddabcf34fc5622',
        'devkit': 'fa75699e90414af021442c21a62c3abf',
    }

    def __init__(self, norm_par: dict[str, list[float]] = {'mean': [0.485, 0.456, 0.406],
                                                           'std': [0.229, 0.224, 0.225], },
                 **kwargs):
        super().__init__(norm_par=norm_par, **kwargs)

    def initialize_folder(self):
        try:
            datasets.ImageNet(root=self.folder_path, split='train')
            datasets.ImageNet(root=self.folder_path, split='val')
        except RuntimeError:
            raise RuntimeError('\n\n'
                               'You need to visit https://image-net.org/download-images.php '
                               'to download ImageNet.\n'
                               'There are direct links to files, but not legal to distribute. '
                               'Please apply for access permission and find links yourself.\n\n'
                               f'folder_path: {self.folder_path}\n'
                               'expected files:\n'
                               '{folder_path}/ILSVRC2012_devkit_t12.tar.gz\n'
                               '{folder_path}/ILSVRC2012_img_train.tar\n'
                               '{folder_path}/ILSVRC2012_img_val.tar\n'
                               '{folder_path}/meta.bin')
        if not os.path.isdir(os.path.join(self.folder_path, 'valid')):
            os.symlink(os.path.join(self.folder_path, 'val'),
                       os.path.join(self.folder_path, 'valid'))

    def _get_org_dataset(self, mode: str, data_format: str = None,
                         **kwargs) -> datasets.DatasetFolder:
        data_format = data_format or self.data_format
        split = 'val' if mode == 'valid' else mode
        return datasets.ImageNet(root=self.folder_path, split=split, **kwargs)


class Sample_ImageNet(ImageNet):

    name: str = 'sample_imagenet'
    num_classes = 10
    url = {}
    org_folder_name = {}

    def initialize_folder(self):
        _dict = Module(self.__dict__)
        _dict.__delattr__('folder_path')
        imagenet = ImageNet(**_dict)
        class_dict: dict = {}
        json_path = os.path.normpath(os.path.join(
            root_dir, 'data', 'sample_imagenet', 'class_dict.json'))
        with open(json_path, 'r', encoding='utf-8') as f:
            class_dict: dict = json.load(f)
        imagenet.sample(child_name=self.name, class_dict=class_dict)



#!/usr/bin/env python3

import torch
import torch.nn as nn
from torch.nn.utils import _stateless

from typing import Iterable


def fim_diag(module: nn.Module, _input: torch.Tensor,
             parameters: Iterable[nn.Parameter] = None
             ) -> list[torch.Tensor]:
    if parameters is None:
        parameters = tuple(module.parameters())
    _output: torch.Tensor = module(_input)  # (N, C)
    with torch.no_grad():
        prob = _output.softmax(dim=1).unsqueeze(-1)  # (N, C, 1)
    log_prob = _output.log_softmax(dim=1)  # (N, C)
    fim_dict: dict[int, list[torch.Tensor]] = {
        i: [] for i in range(len(parameters))}
    N, C = log_prob.shape
    for n in range(N):
        for c in range(C):
            grad_list = torch.autograd.grad(
                log_prob[n][c], parameters, retain_graph=True)
            for i, grad in enumerate(grad_list):    # different layers
                fim = grad.flatten().square()    # (D)
                fim_dict[i].append(fim.detach().clone())
    fim_list: list[torch.Tensor] = []
    for i, value in fim_dict.items():    # different layers
        D = value[0].shape[0]
        fim = torch.stack(value).view(N, C, D) * prob    # (N, C, D)
        fim = fim.sum(1).mean(0)   # (D)
        fim_list.append(fim)
    return fim_list


def old_fim(module: nn.Module, _input: torch.Tensor,
            parameters: Iterable[nn.Parameter] = None
            ) -> list[torch.Tensor]:
    if parameters is None:
        parameters = tuple(module.parameters())
    _output: torch.Tensor = module(_input)  # (N, C)
    with torch.no_grad():
        prob = _output.softmax(dim=1).unsqueeze(-1).unsqueeze(-1)  # (N, C, 1, 1)
    log_prob = _output.log_softmax(dim=1)  # (N, C)
    fim_dict: dict[int, list[torch.Tensor]] = {
        i: [] for i in range(len(parameters))}
    N, C = log_prob.shape
    for n in range(N):
        for c in range(C):
            grad_list = torch.autograd.grad(
                log_prob[n][c], parameters, retain_graph=True)
            for i, grad in enumerate(grad_list):    # different layers
                flatten_grad = grad.flatten()    # (D)
                fim = flatten_grad.outer(flatten_grad)   # (D, D)
                fim_dict[i].append(fim.detach().clone())
    fim_list: list[torch.Tensor] = []
    for i, value in fim_dict.items():    # different layers
        D = value[0].shape[0]
        fim = torch.stack(value).view(N, C, D, D) * prob    # (N, C, D, D)
        fim_list.append(fim.sum(1).mean(0))   # (D, D)
    return fim_list


def new_fim(module: nn.Module, _input: torch.Tensor,
            parameters: dict[str, nn.Parameter] | Iterable[nn.Parameter] = None
            ) -> list[torch.Tensor]:
    if not isinstance(parameters, dict):
        id_map: dict[torch.Tensor, str] = {
            param.data: name for name, param in module.named_parameters()}
        parameters = {id_map[param.data if isinstance(
            param, torch.nn.Parameter) else param]: param
            for param in parameters}
    if parameters is None:
        parameters = dict(module.named_parameters())
    keys, values = zip(*parameters.items())

    with torch.no_grad():
        _output: torch.Tensor = module(_input)  # (N, C)
        prob = _output.softmax(dim=1).unsqueeze(-1).unsqueeze(-1)  # (N, C, 1, 1)

    def func(*params: torch.Tensor):
        _output: torch.Tensor = _stateless.functional_call(
            module, {n: p for n, p in zip(keys, params)}, _input)
        return _output.log_softmax(dim=1)  # (N, C)
    jac: tuple[torch.Tensor] = torch.autograd.functional.jacobian(
        func, values)
    jac = (j.flatten(2) for j in jac)   # (N, C, D)

    fim_list: list[torch.Tensor] = []
    for j in jac:  # TODO: parallel
        fim = prob * j.unsqueeze(-1) * \
            j.unsqueeze(-2)  # (N, C, D, D)
        fim_list.append(fim.sum(1).mean(0))   # (D, D)
    return fim_list



#!/usr/bin/env python3

from trojanvision.environ import env
from trojanzoo.defenses import Defense
from trojanzoo.utils.logger import MetricLogger
from trojanzoo.utils.metric import mask_jaccard, normalize_mad
from trojanzoo.utils.output import output_iter, prints
from trojanzoo.utils.tensor import tanh_func
from trojanzoo.utils.data import TensorListDataset, sample_batch

import torch
import torch.optim as optim
import numpy as np
from sklearn import metrics

import os
from abc import abstractmethod

from typing import TYPE_CHECKING
from trojanvision.datasets import ImageSet
from trojanvision.models import ImageModel
from trojanvision.attacks.backdoor import BadNet
import argparse
from collections.abc import Iterable
if TYPE_CHECKING:
    import torch.utils.data    # TODO: python 3.10


def format_list(_list: list, _format: str = ':8.3f') -> str:
    return '[' + ', '.join(['{{{}}}'.format(_format).format(a) for a in _list]) + ']'


class BackdoorDefense(Defense):
    r"""Backdoor defense abstract class.
    It inherits :class:`trojanzoo.defenses.Defense`.

    Args:
        original (bool): Whether to load original clean model.
            If ``False``, load attack poisoned model
            by calling ``self.attack.load()``.

    Attributes:
        real_mark (torch.Tensor): Watermark that the attacker uses
            with shape ``(C+1, H, W)``.
        real_mask (torch.Tensor): Mask of the watermark
            by calling :meth:`trojanvision.marks.Watermark.get_mask()`.
    """
    name: str = 'backdoor_defense'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--original', action='store_true',
                           help='whether to load original clean model '
                           '(default: False)')
        return group

    def __init__(self, attack: BadNet, original: bool = False, **kwargs):
        self.original: bool = original
        if not self.original:
            attack.load(**kwargs)
        super().__init__(attack=attack, **kwargs)
        self.dataset: ImageSet
        self.model: ImageModel
        self.attack: BadNet
        self.real_mark = self.attack.mark.mark.clone()
        self.real_mask = self.attack.mark.get_mask()

    @abstractmethod
    def detect(self, **kwargs):
        self.attack.validate_fn()

    def get_filename(self, **kwargs):
        r"""Get filenames for current defense settings."""
        return self.attack.name + '_' + self.attack.get_filename(**kwargs)


class InputFiltering(BackdoorDefense):
    r"""Backdoor defense abstract class of input filtering.
    It inherits :class:`trojanvision.defenses.BackdoorDefense`.

    It detects whether a test input is poisoned.

    The defense tests :attr:`defense_input_num` clean test inputs
    and their corresponding poison version
    (``2 * defense_input_num`` in total).

    Args:
        defense_input_num (int): Number of test inputs.
            Defaults to ``100``.

    Attributes:
        test_set (torch.utils.data.Dataset): Test dataset
            with length :attr:`defense_input_num`.
    """
    name: str = 'input_filtering'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--defense_input_num', type=int,
                           help='number of test inputs (default: 100)')
        return group

    def __init__(self, defense_input_num: int = 100, **kwargs):
        super().__init__(**kwargs)
        self.param_list['input_filtering'] = ['defense_input_num']
        self.defense_input_num = defense_input_num
        self.test_input, self.test_label = self.get_test_data()

    def detect(self, **kwargs):
        super().detect(**kwargs)
        y_true = self.get_true_labels()
        y_pred = self.get_pred_labels()
        tn, fp, fn, tp = metrics.confusion_matrix(y_true, y_pred).ravel()
        print()
        print(f'{tn=:d} {fp=:d} {fn=:d} {tp=:d}')
        print(f'f1_score        : {metrics.f1_score(y_true, y_pred):8.3f}')
        print(f'precision_score : {metrics.precision_score(y_true, y_pred):8.3f}')
        print(f'recall_score    : {metrics.recall_score(y_true, y_pred):8.3f}')
        print(f'accuracy_score  : {metrics.accuracy_score(y_true, y_pred):8.3f}')

    def get_test_data(self) -> tuple[torch.Tensor, torch.Tensor]:
        r"""Get test data.

        Returns:
            (torch.Tensor, torch.Tensor):
                Input and label tensors
                with length ``defense_input_num``.
        """
        input_list = []
        label_list = []
        remain_counter = self.defense_input_num
        for data in self.dataset.loader['valid']:
            _input, _label = self.model.remove_misclassify(data)
            if len(_label) == 0:
                continue
            trigger_input = self.attack.add_mark(_input)
            trigger_label = self.attack.target_class * torch.ones_like(_label)
            _classification = self.model.get_class(trigger_input)
            repeat_idx = _classification.eq(trigger_label)
            _input, _label = _input[repeat_idx], _label[repeat_idx]
            if len(_label) == 0:
                continue
            if len(_input) < remain_counter:
                remain_counter -= len(_input)
            else:
                _input = _input[:remain_counter]
                _label = _label[:remain_counter]
                remain_counter = 0
            input_list.append(_input.cpu())
            label_list.extend(_label.cpu().tolist())
            if remain_counter == 0:
                break
        else:
            raise Exception('No enough test data')
        return torch.cat(input_list), label_list

    def get_true_labels(self) -> torch.Tensor:
        r"""Get ground-truth labels for test inputs.

        Defaults to return ``[False] * defense_input_num + [True] * defense_input_num``.

        Returns:
            torch.Tensor: ``torch.BoolTensor`` with shape ``(2 * defense_input_num)``.
        """
        zeros = torch.zeros(self.defense_input_num, dtype=torch.bool)
        ones = torch.ones_like(zeros)
        return torch.cat([zeros, ones])

    def get_pred_labels(self) -> torch.Tensor:
        r"""Get predicted labels for test inputs (need overriding).

        Returns:
            torch.Tensor: ``torch.BoolTensor`` with shape ``(2 * defense_input_num)``.
        """
        ...


class TrainingFiltering(BackdoorDefense):
    r"""Backdoor defense abstract class of training data filtering.
    It inherits :class:`trojanvision.defenses.BackdoorDefense`.

    Provided :attr:`defense_input_num` training data,
    it detects which training data is poisoned.

    The defense evaluates clean and poison training inputs.

    - If :attr:`defense_input_num` is ``None``, use full training data.
    - Else, sample ``defense_input_num * poison_percent`` poison training data
      and ``defense_input_num * (1 - poison_percent)`` clean training data.

    If dataset is not using ``train_mode == 'dataset'``,
    construct poison dataset using all clean data with watermark attached.
    (If :attr:`defense_input_num` is ``None`` as well,
    the defense will evaluate the whole clean training set and its poisoned version.)

    Args:
        defense_input_num (int): Number of training inputs to evaluate.
            Defaults to ``None`` (all training set).

    Attributes:
        clean_set (torch.utils.data.Dataset): Clean training data to evaluate.
        poison_set (torch.utils.data.Dataset): Poison training data to evaluate.
    """
    name: str = 'training_filtering'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--defense_input_num', type=int,
                           help='the number of training inputs to evaluate '
                           '(default: None)')
        return group

    def __init__(self, defense_input_num: int = None, **kwargs):
        super().__init__(**kwargs)
        self.defense_input_num = defense_input_num
        self.clean_set, self.poison_set = self.get_datasets()

    def get_datasets(self) -> tuple[torch.utils.data.Dataset, torch.utils.data.Dataset]:
        r"""Get clean and poison datasets.

        Returns:
            (torch.utils.data.Dataset, torch.utils.data.Dataset):
                Clean training dataset and poison training dataset.
        """
        if self.attack.poison_set is None:
            self.attack.poison_set = self.attack.get_poison_dataset(
                poison_num=len(self.dataset.loader['train'].dataset))
        if not self.defense_input_num:
            return self.dataset.loader['train'].dataset, self.attack.poison_set
        if self.attack.train_mode != 'dataset':
            poison_num = int(self.defense_input_num * self.attack.poison_percent)
            clean_num = self.defense_input_num - poison_num
            clean_input, clean_label = sample_batch(self.dataset.loader['train'].dataset,
                                                    batch_size=clean_num)
            trigger_input, trigger_label = sample_batch(self.attack.poison_set,
                                                        batch_size=poison_num)
            clean_set = TensorListDataset(clean_input, clean_label.tolist())
            poison_set = TensorListDataset(trigger_input, trigger_label.tolist())
        return clean_set, poison_set

    def detect(self, **kwargs):
        super().detect(**kwargs)
        y_pred = self.get_pred_labels()
        y_true = self.get_true_labels()
        print(f'f1_score        : {metrics.f1_score(y_true, y_pred):8.3f}')
        print(f'precision_score : {metrics.precision_score(y_true, y_pred):8.3f}')
        print(f'recall_score    : {metrics.recall_score(y_true, y_pred):8.3f}')
        print(f'accuracy_score  : {metrics.accuracy_score(y_true, y_pred):8.3f}')

    def get_true_labels(self) -> torch.Tensor:
        r"""Get ground-truth labels for training inputs.

        Defaults to return ``[False] * len(self.clean_set) + [True] * len(self.poison_set)``.

        Returns:
            torch.Tensor: ``torch.BoolTensor`` with shape ``(defense_input_num)``.
        """
        return torch.cat([torch.zeros(len(self.clean_set), dtype=torch.bool),
                          torch.ones(len(self.poison_set), dtype=torch.bool)])

    @abstractmethod
    def get_pred_labels(self) -> torch.Tensor:
        r"""Get predicted labels for training inputs (need overriding).

        Returns:
            torch.Tensor: ``torch.BoolTensor`` with shape ``(defense_input_num)``.
        """
        ...


class ModelInspection(BackdoorDefense):
    r"""Backdoor defense abstract class of model inspection.
    It inherits :class:`trojanvision.defenses.BackdoorDefense`.

    Provided a model, it tries to search for a trigger.
    If trigger exists, that means the model is poisoned.

    Args:
        defense_remask_epoch (int): Defense watermark optimizing epochs.
            Defaults to ``10``.
        defense_remask_lr (float): Defense watermark optimizing learning rate.
            Defaults to ``0.1``.
        cost (float): Cost of mask norm loss.
            Defaults to ``1e-3``.

    Attributes:
        cost (float): Cost of mask norm loss.
        clean_set (torch.utils.data.Dataset): Clean training data to evaluate.
        poison_set (torch.utils.data.Dataset): Poison training data to evaluate.
    """
    name: str = 'model_inspection'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--defense_remask_epoch', type=int,
                           help='defense watermark optimizing epochs '
                           '(default: 10)')
        group.add_argument('--defense_remask_lr', type=float,
                           help='defense watermark optimizing learning rate '
                           '(default: 0.1)')
        group.add_argument('--cost', type=float,
                           help='cost of mask norm loss '
                           '(default: 1e-3)')
        return group

    def __init__(self, defense_remask_epoch: int = 10,
                 defense_remask_lr: float = 0.1,
                 cost: float = 1e-3, **kwargs):
        super().__init__(**kwargs)
        self.param_list['model_inspection'] = ['defense_remask_epoch',
                                               'defense_remask_lr',
                                               'cost']

        self.defense_remask_epoch = defense_remask_epoch
        self.defense_remask_lr = defense_remask_lr
        self.cost = cost

    def detect(self, **kwargs):
        super().detect(**kwargs)

        self.mark_random_pos = self.attack.mark.mark_random_pos
        mark_keys = ['mark', 'mark_height', 'mark_width',
                     'mark_height_offset', 'mark_width_offset',
                     'mark_random_pos', ]
        self.mark_dict = {key: getattr(self.attack.mark, key) for key in mark_keys}
        self.new_dict = {'mark': torch.zeros(self.attack.mark.mark.size(0),
                                             self.attack.mark.data_shape[-2],
                                             self.attack.mark.data_shape[-1],
                                             device=self.attack.mark.mark.device),
                         'mark_height': self.attack.mark.data_shape[-2],
                         'mark_width': self.attack.mark.data_shape[-1],
                         'mark_height_offset': 0,
                         'mark_width_offset': 0,
                         'mark_random_pos': False,
                         }

        for k, v in self.new_dict.items():
            setattr(self.attack.mark, k, v)
        self.attack.mark.mark.zero_()

        mark_list, loss_list, asr_list = self.get_mark_loss_list()
        mask_norms: torch.Tensor = mark_list[:, -1].flatten(start_dim=1).norm(p=1, dim=1)
        mask_norm_list: list[float] = mask_norms.tolist()
        print()
        print('asr           : ' + format_list(asr_list))
        print('mask norms    : ' + format_list(mask_norm_list))
        print('loss          : ' + format_list(loss_list))
        print()
        print('asr MAD       : ' + format_list(normalize_mad(asr_list).tolist()))
        print('mask norm MAD : ' + format_list(normalize_mad(mask_norms).tolist()))
        print('loss MAD      : ' + format_list(normalize_mad(loss_list).tolist()))

        if not self.mark_random_pos:
            self.attack.mark.mark = mark_list[self.attack.target_class]
            select_num = self.attack.mark.mark_height * self.attack.mark.mark_width
            overlap = mask_jaccard(self.attack.mark.get_mask(),
                                   self.real_mask,
                                   select_num=select_num)
            print(f'Jaccard index: {overlap:.3f}')

    def get_mark_loss_list(self, verbose: bool = True,
                           **kwargs) -> tuple[torch.Tensor, list[float], list[float]]:
        r"""Get list of mark, loss, asr of recovered trigger for each class.

        Args:
            verbose (bool): Whether to output jaccard index for each trigger.
                It's also passed to :meth:`optimize_mark()`.
            **kwargs: Keyword arguments passed to :meth:`optimize_mark()`.

        Returns:
            (torch.Tensor, list[float], list[float]):
                list of mark, loss, asr with length ``num_classes``.
        """
        mark_list: list[torch.Tensor] = []
        loss_list: list[float] = []
        asr_list: list[float] = []
        # todo: parallel to avoid for loop
        file_path = os.path.normpath(os.path.join(
            self.folder_path, self.get_filename() + '.npz'))
        for label in range(self.model.num_classes):
            print('Class: ', output_iter(label, self.model.num_classes))
            mark, loss = self.optimize_mark(label, verbose=verbose, **kwargs)
            if verbose:
                asr, _ = self.attack.validate_fn(indent=4)
                if not self.mark_random_pos:
                    select_num = self.attack.mark.mark_height * self.attack.mark.mark_width
                    overlap = mask_jaccard(self.attack.mark.get_mask(),
                                           self.real_mask,
                                           select_num=select_num)
                    prints(f'Jaccard index: {overlap:.3f}', indent=4)
            else:
                asr, _ = self.model._validate(get_data_fn=self.attack.get_data,
                                              keep_org=False, poison_label=True,
                                              verbose=False)
            mark_list.append(mark)
            loss_list.append(loss)
            asr_list.append(asr)
            np.savez(file_path, mark_list=np.stack([mark.detach().cpu().numpy() for mark in mark_list]),
                     loss_list=np.array(loss_list))
        print()
        print('Defense results saved at: ' + file_path)
        mark_list_tensor = torch.stack(mark_list)
        return mark_list_tensor, loss_list, asr_list

    def loss(self, _input: torch.Tensor, _label: torch.Tensor, target: int,
             trigger_output: None | torch.Tensor = None,
             **kwargs) -> torch.Tensor:
        r"""Loss function to optimize recovered trigger.

        Args:
            _input (torch.Tensor): Clean input tensor
                with shape ``(N, C, H, W)``.
            _label (torch.Tensor): Clean label tensor
                with shape ``(N)``.
            target (int): Target class.
            trigger_output (torch.Tensor):
                Output tensor of input tensor with trigger.
                Defaults to ``None``.

        Returns:
            torch.Tensor: Scalar loss tensor.
        """
        trigger_input = self.attack.add_mark(_input)
        trigger_label = target * torch.ones_like(_label)
        if trigger_output is None:
            trigger_output = self.model(trigger_input, **kwargs)
        return self.model.loss(trigger_input, trigger_label, _output=trigger_output)

    def optimize_mark(self, label: int,
                      loader: Iterable = None,
                      logger_header: str = '',
                      verbose: bool = True,
                      **kwargs) -> tuple[torch.Tensor, float]:
        r"""
        Args:
            label (int): The class label to optimize.
            loader (collections.abc.Iterable):
                Data loader to optimize trigger.
                Defaults to ``self.dataset.loader['train']``.
            logger_header (str): Header string of logger.
                Defaults to ``''``.
            verbose (bool): Whether to use logger for output.
                Defaults to ``True``.
            **kwargs: Keyword arguments passed to :meth:`loss()`.

        Returns:
            (torch.Tensor, torch.Tensor):
                Optimized mark tensor with shape ``(C + 1, H, W)``
                and loss tensor.
        """
        atanh_mark = torch.randn_like(self.attack.mark.mark, requires_grad=True)
        optimizer = optim.Adam([atanh_mark], lr=self.defense_remask_lr, betas=(0.5, 0.9))
        lr_scheduler = optim.lr_scheduler.CosineAnnealingLR(optimizer,
                                                            T_max=self.defense_remask_epoch)
        optimizer.zero_grad()
        loader = loader or self.dataset.loader['train']

        # best optimization results
        norm_best: float = float('inf')
        mark_best: torch.Tensor = None
        loss_best: float = None

        logger = MetricLogger(indent=4)
        logger.create_meters(loss='{last_value:.3f}',
                             acc='{last_value:.3f}',
                             norm='{last_value:.3f}',
                             entropy='{last_value:.3f}',)
        batch_logger = MetricLogger()
        logger.create_meters(loss=None, acc=None, entropy=None)

        iterator = range(self.defense_remask_epoch)
        if verbose:
            iterator = logger.log_every(iterator, header=logger_header)
        for _ in iterator:
            batch_logger.reset()
            for data in loader:
                self.attack.mark.mark = tanh_func(atanh_mark)    # (c+1, h, w)
                _input, _label = self.model.get_data(data)
                trigger_input = self.attack.add_mark(_input)
                trigger_label = label * torch.ones_like(_label)
                trigger_output = self.model(trigger_input)

                batch_acc = trigger_label.eq(trigger_output.argmax(1)).float().mean()
                batch_entropy = self.loss(_input, _label,
                                          target=label,
                                          trigger_output=trigger_output,
                                          **kwargs)
                batch_norm: torch.Tensor = self.attack.mark.mark[-1].norm(p=1)
                batch_loss = batch_entropy + self.cost * batch_norm

                batch_loss.backward()
                optimizer.step()
                optimizer.zero_grad()

                batch_size = _label.size(0)
                batch_logger.update(n=batch_size,
                                    loss=batch_loss.item(),
                                    acc=batch_acc.item(),
                                    entropy=batch_entropy.item())
            lr_scheduler.step()
            self.attack.mark.mark = tanh_func(atanh_mark)    # (c+1, h, w)

            # check to save best mask or not
            loss = batch_logger.meters['loss'].global_avg
            acc = batch_logger.meters['acc'].global_avg
            norm = float(self.attack.mark.mark[-1].norm(p=1))
            entropy = batch_logger.meters['entropy'].global_avg
            if norm < norm_best:
                mark_best = self.attack.mark.mark.detach().clone()
                loss_best = loss
                logger.update(loss=loss, acc=acc, norm=norm, entropy=entropy)

            if self.check_early_stop(loss=loss, acc=acc, norm=norm, entropy=entropy):
                print('early stop')
                break
        atanh_mark.requires_grad_(False)
        self.attack.mark.mark = mark_best
        return mark_best, loss_best

    def check_early_stop(self, *args, **kwargs) -> bool:
        r"""Check whether to early stop at the end of each remask epoch.

        Returns:
            bool: Whether to early stop. Defaults to ``False``.
        """
        return False

    def load(self, path: None | str = None):
        r"""Load recovered mark from :attr:`path`.

        Args:
            path (str): npz path of recovered mark.
                Defaults to ``'{folder_path}/{self.get_filename()}.npz'``.
        """
        if path is None:
            path = os.path.join(self.folder_path, self.get_filename() + '.npz')
        _dict = np.load(path)
        for k, v in self.new_dict.items():
            setattr(self.attack.mark, k, v)
        self.attack.mark.mark = torch.from_numpy(_dict['mark_list'][self.attack.target_class]).to(device=env['device'])
        print('defense results loaded from:', path)



#!/usr/bin/env python3

from .param import Module, Param
from .process import BasicObject, Process, ModelProcess
from trojanzoo.utils.output import ansi, prints

import sys


def get_name(name: str = None, module: str | object = None,
             arg_list: list[str] = []) -> str:
    if module is not None:
        if isinstance(module, str):
            return module
        try:
            return getattr(module, 'name')
        except AttributeError:
            raise TypeError(f'{type(module)}    {module}')
    if name is not None:
        return name
    argv = sys.argv
    for arg in arg_list:
        try:
            idx = argv.index(arg)
            name: str = argv[idx + 1]
        except ValueError:
            continue
    return name


def summary(indent: int = 0, **kwargs):
    for key, value in kwargs.items():
        prints('{yellow}{0:<10s}{reset}'.format(key, **ansi),
               indent=indent)
        try:
            value.summary()
        except AttributeError:
            prints(value, indent=10)
        prints('-' * 30, indent=indent)
        print()
